import math
import sys


def Hellinger(parent, children):
    if len(children) != 2:
        raise Exception(
            "Hellinger Distance needs only two child nodes (binary split)")

    if sum(parent) == max(parent):
        return 0
    if (parent[0] != 0 and parent[1] != 0):
        s1p = math.sqrt(children[0][0] / parent[0])
        s1n = math.sqrt(children[0][1] / parent[1])

        s2p = math.sqrt(children[1][0] / parent[0])
        s2n = math.sqrt(children[1][1] / parent[1])

        result = math.sqrt(math.pow(s1p - s1n, 2) + math.pow(s2p - s2n, 2))

        return result
    else:
         return 0


def MultiClassHellinger(parent, children):
    if len(children) != 2:
        raise Exception(
            "Hellinger Distance needs only two child nodes (binary split)")

    if sum(parent) == max(parent):
        return 0

    hellinger = sys.float_info.min

    try:
        for i in range(len(parent)):
            tn = __SumDifferent(parent, i)

            if parent[i] == 0.0:
                if children[0][i] == 0.0:
                    s1p = float('nan')
                else: s1p = float('inf')
            else: s1p = math.sqrt(children[0][i] / parent[i])

            s1nA = __SumDifferent(children[0], i)

            if (tn == 0.0):
                if s1nA == 0.0:
                    s1n = float('nan')
                else:
                    s1n = float('inf')
            else: s1n = math.sqrt(s1nA / tn)

            if parent[i] == 0.0:
                if children[1][i] == 0.0:
                    s2p = float('nan')
                else: s2p = float('inf')
            else: s2p = math.sqrt(children[1][i] / parent[i])

            s2nA = __SumDifferent(children[1], i)
            if (tn == 0.0):
                if s2nA == 0.0:
                    s2n = float('nan')
                else:
                    s2n = float('inf')
            else: s2n = math.sqrt(s2nA / tn)

            currentValue = math.pow(s1p - s1n, 2) + math.pow(s2p - s2n, 2)
            
            if currentValue > hellinger:
                hellinger = currentValue
    except ZeroDivisionError:
        raise Exception(f"Division by Zero ocurrend")

    if hellinger == float('inf'):
        return sys.float_info.max

    return math.sqrt(hellinger)

def QuinlanGain (parent, children):
    result = __GetImpurity(parent)
    total = sum(parent)
    nonMissing = 0.0

    for distribution in children:
        childCount = sum(distribution)
        nonMissing += childCount
        result -= __GetImpurity(distribution) * (childCount * 1.0 / total)

    return result * (nonMissing) / total 


def __SumDifferent(vector, index):
    sumValue = 0
    for i in range(len(vector)):
        if index != i:
            sumValue += vector[i]
    return sumValue

def __GetImpurity (distribution) :
    result = 0
    count = sum(distribution)

    for value in distribution:
        if (value != 0):
            p = value * 1.0 / count
            result -= p * math.log(p, 2)

    return result



import os
import argparse
import numpy as np
import pandas as pd

from tqdm import tqdm, trange
from core.PBC4cip import PBC4cip
from core.FileManipulation import WritePatternsBinary, WritePatternsCSV, WriteResultsCSV
from core.Evaluation import obtainAUCMulticlass
from core.Helpers import get_col_dist, get_idx_val
from datetime import datetime

def CheckSuffix(file, suffix):
    if not suffix or len(suffix) == 0:
        return True
    if not suffix in file or len(suffix) >= len(file):
        return False
    filename, file_extension = os.path.splitext(file)
    return filename[(len(suffix)*-1):] == suffix


def GetFilesFromDirectory(directory):
    print(directory)
    files = []
    if os.path.isdir(directory):
        for r, d, f in os.walk(directory):
            for file in f:
                files.append(os.path.join(r, file))
        return files
    else:
        raise Exception(f"Directory {directory} is not valid.")


def import_data(trainFile, testFile):
    train = pd.read_csv(trainFile, sep= ',') 
    test = pd.read_csv(testFile, sep= ',')

    return train, test

def split_data(train, test):
    X_train = train.iloc[:,  0:train.shape[1]-1]
    y_train =  train.iloc[:, train.shape[1]-1 : train.shape[1]]

    X_test = test.iloc[:,  0:test.shape[1]-1]
    y_test =  test.iloc[:, test.shape[1]-1 : test.shape[1]]

    return X_train, y_train, X_test, y_test

def score(predicted, y):
        y_class_dist = get_col_dist(y[f'{y.columns[0]}'])
        real = list(map(lambda instance: get_idx_val(y_class_dist, instance), y[f'{y.columns[0]}']))
        numClasses = len(y_class_dist)
        confusion = [[0]*2 for i in range(numClasses)]
        classified_as = 0
        error_count = 0

        for i in range(len(real)):
            if real[i] != predicted[i]:
                error_count = error_count + 1
            confusion[real[i]][predicted[i]] = confusion[real[i]][predicted[i]] + 1

        acc = 100.0 * (len(real) - error_count) / len(real)
        auc = obtainAUCMulticlass(confusion, numClasses)

        return confusion, acc, auc


def test_PBC4cip(trainFile, testFile):
    train, test = import_data(trainFile, testFile)
    X_train, y_train, X_test, y_test = split_data(train, test)
    
    classifier = PBC4cip()
    patterns = classifier.fit(X_train, y_train)

    y_test_scores = classifier.score_samples(X_test)
    print(f"Test Scores: {y_test_scores}")
    
    y_pred = classifier.predict(y_test_scores)
    confusion, acc, auc = score(y_pred, y_test)

    
    print(f"Patterns Found: ")
    for pattern in patterns:
        print(f"{pattern}")

    print(f"\nConfusion Matrix:")
    for i in range(len(confusion[0])):
        for j in range(len(confusion[0])):
            print(f"{confusion[i][j]} ", end='')
        print("")
    print(f"\n\nacc: {acc} , auc: {auc} , numPatterns: {len(patterns)}")
    

def Execute(args):
    testing_files = []
    training_files = []

    if args.training_files:
        training_files = args.training_files

    if args.training_directory:
        training_files = list(set().union(training_files, list(filter(
            lambda file: CheckSuffix(file, args.training_file_suffix), GetFilesFromDirectory(args.training_directory)))))

    if args.input_files:
        testing_files = args.input_files

    if args.input_directory:
        testing_files = list(set().union(testing_files, list(filter(
            lambda file: CheckSuffix(file, args.test_file_suffix), GetFilesFromDirectory(args.input_directory)))))

    training_files.sort()
    testing_files.sort()

    print(
        f"Training files detected (*{args.training_file_suffix}.[arff|dat]): {len(training_files)}")
    print(
        f"Testing files detected (*{args.test_file_suffix}.[arff|dat]): {len(testing_files)}")

    print("===============================================================================")
    tra = trange(len(training_files), desc='Training and Testing Files...', leave=True, unit="dataset")

    now = datetime.now()
    resultsId = now.strftime("%Y%m%d%H%M%S")
    for f in tra:
        tra.set_description(f"Working from {training_files[f]}")
        test_PBC4cip(training_files[f], testing_files[f] )



if __name__ == '__main__':

    defaultDataDir = os.path.join(os.path.normpath(
        os.path.join(os.getcwd(), os.pardir)), "data", "python")
    defaultOutputDir = os.path.join(os.path.normpath(
        os.path.join(os.getcwd(), os.pardir)), "output")

    defaultTrainingFiles = list()
    defaultTestingFiles = list()
    parser = argparse.ArgumentParser(
        description="Process class imbalanced datasets using PBC4cip.")

    parser.add_argument("--training-files",
                        type=str,
                        metavar="<*.dat/*.arff>",
                        nargs="+",
                        help="a file or files that are going to be used to train the classifier")

    parser.add_argument("--training-directory",
                        type=str,
                        metavar="'"+defaultDataDir+"'",
                        help="the directory with files to be used to train the classifier")

    parser.add_argument("--input-files",
                        type=str,
                        metavar="<*.dat/*.arff>",
                        nargs="+",
                        help="a file or files to be classified")
    
    parser.add_argument("--input-directory",
                        type=str,
                        metavar="'"+defaultDataDir+"'",
                        help="the directory with files to be classified")
    
    parser.add_argument("--output-directory",
                        type=str,
                        metavar="'"+defaultOutputDir+"'",
                        default=defaultOutputDir,
                        help="the output directory for the patterns")
    
    parser.add_argument("--test-file-suffix",
                        type=str,
                        metavar="'tst'",
                        default="tst",
                        help="states which suffix will indicate the test files")

    parser.add_argument("--training-file-suffix",
                        type=str,
                        metavar="'tra'",
                        default="tra",
                        help="states which suffix will indicate the training files")

    args = parser.parse_args()
    if not args.training_files and not args.training_directory and not args.input_files and not args.input_directory:
        parser.print_help()
    else:
        Execute(args)



import math

class SingleFeatureSelector(object):
    def __init__(self, dataset, feature):
        self.__ChildrenCount = 2
        self.__Dataset = dataset
        self.__Model = dataset.Model
        self.__Feature = feature

    @property
    def ChildrenCount(self):
        return self.__ChildrenCount
    @ChildrenCount.setter
    def ChildrenCount(self, new_children_count):
        self.__ChildrenCount = new_children_count
    
    @property
    def Dataset(self):
        return self.__Dataset
    @Dataset.setter
    def Dataset(self, new_dataset):
        self.__Dataset = new_dataset

    @property
    def Feature(self):
        return self.__Feature
    @Feature.setter
    def Feature(self, new_feature):
        self.__Feature = new_feature


    def Select(self, instance):
        if self.Dataset.IsMissing(self.Feature, instance):
            return None

    def __format__(self, index):
        return self.__repr__()

    def __repr__(self):
        return f"¿[{self.Feature[0]}]?"


class CutPointSelector(SingleFeatureSelector):
    def __init__(self, dataset, feature):
        super().__init__(dataset, feature)
        self.__CutPoint = None

    def Select(self, instance):
        super().Select(instance)
        if self.Dataset.IsNominalFeature(self.Feature):
            raise Exception("Cannot use cutpoint on nominal data")
        if self.Dataset.GetFeatureValue(self.Feature, instance) <= self.CutPoint:
            return [1.0, 0.0]
        else:
            return [0.0, 1.0]

    def __format__(self, index):
        if index == 0:
            return f"{self.Feature[0]}<={self.CutPoint}"
        else:
            return f"{self.Feature[0]}>{self.CutPoint}"

    def __repr__(self):
        return f"{self.Feature[0]}<={self.CutPoint}"


class MultipleValuesSelector(SingleFeatureSelector):
    def __init__(self, dataset, feature):
        super().__init__(dataset, feature)
        self.__Values = []

    def Select(self, instance):
        super().Select(instance)
        if not self.Dataset.IsNominalFeature(self.Feature):
            raise Exception("Cannot use multiple values on non-nominal data")
        value = self.Dataset.GetFeatureValue(self.Feature, instance)
        index = self.Values.index(value)
        if index == -1:
            return None
        result = [0]*self.ChildrenCount
        result[index] = 1
        return result

    def __format__(self, index):
        return f"{self.Feature[0]}={self.Dataset.GetValueOfIndex(self.Feature[0],index)}"

    def __repr__(self):
        return f"{self.Feature[0]}in[{', '.join(map(lambda value: self.Dataset.GetValueOfIndex(self.Feature[0],value),self.Values))}]"


class ValueAndComplementSelector(SingleFeatureSelector):
    def __init__(self, dataset, feature):
        super().__init__(dataset, feature)
        self.__Value = None

    def Select(self, instance):
        super().Select(instance)
        if not self.Dataset.IsNominalFeature(self.Feature):
            raise Exception("Cannot use multiple values on non-nominal data")
        if self.Dataset.GetFeatureValue(self.Feature, instance) == self.Value:
            return [1, 0]
        else:
            return [0, 1]

    def __format__(self, index):
        if index == 0:
            return f"{self.Feature[0]}={self.Dataset.GetValueOfIndex(self.Feature[0],self.Value)}"
        else:
            return f"{self.Feature[0]}<>{self.Dataset.GetValueOfIndex(self.Feature[0],self.Value)}"

    def __repr__(self):
        return f"{self.Feature[0]}={self.Dataset.GetValueOfIndex(self.Feature[0],self.Value)}"


class MultipleFeaturesSelector(object):
    def __init__(self, dataset, features):
        self.ChildrenCount = 2
        self.Dataset = dataset
        self.Model = dataset.Model
        self.Features = features

    def Select(self, instance):
        if any(self.Dataset.IsMissing(feature, instance) for feature in self.Features):
            return None

    def __format__(self, index):
        return self.__repr__()

    def __repr__(self):
        linearCombination = ' + '.join(
            map(lambda feature: "1.0 * " + feature[0], self.Features))
        return f"¿[{linearCombination}]?"


class MultivariateCutPointSelector(MultipleFeaturesSelector):
    def __init__(self, dataset, feature):
        super().__init__(dataset, feature)
        self.__CutPoint = None
        self.__Weights = None

    def Select(self, instance):
        super().Select(instance)
        if any(self.Dataset.IsNominalFeature(feature) for feature in self.Features):
            raise Exception("Cannot use cutpoint on nominal data")
        if self.Dataset.ScalarProjection(
                instance, self.Features, self.Weights) <= self.CutPoint:
            return [1, 0]
        else:
            return [0, 1]

    def __format__(self, index):
        linearCombination = ' + '.join(
            map(lambda weight: str(self.Weights[weight]) + " * " + weight[0], self.Weights))
        if index == 0:
            return f"{linearCombination}<={self.CutPoint}"
        else:
            return f"{linearCombination}>{self.CutPoint}"

    def __repr__(self):
        linearCombination = ' + '.join(
            map(lambda weight: str(self.Weights[weight]) + " * " + weight[0], self.Weights))
        return f"{linearCombination}<={self.CutPoint}"




from .FeatureSelectors import SingleFeatureSelector, CutPointSelector, MultipleValuesSelector, ValueAndComplementSelector, MultivariateCutPointSelector
from .Dataset import Dataset
import math


class SubsetRelation(object):
    Unknown = 0
    Unrelated = 1
    Equal = 2
    Subset = 3
    Superset = 4
    Different = 5


class Item(object):

    def __init__(self, dataset, feature):
        self.__Dataset = dataset
        self.__Feature = feature
    
    @property
    def Dataset(self):
        return self.__Dataset
    @Dataset.setter
    def Dataset(self, new_dataset):
        self.__Dataset = new_dataset
    
    @property
    def Feature(self):
        return self.__Feature
    @Feature.setter
    def Feature(self, new_feature):
        self.__Feature = new_feature

    def IsMatch(self, instance):
        return False

    def CompareTo(self, other):
        return SubsetRelation.Unknown


class SingleValueItem(Item):

    def __init__(self, dataset, feature, value):
        super().__init__(dataset, feature)
        self.Value = value

    def IsMatch(self, instance):
        return not self.Dataset.IsMissing(self.Feature, instance)

    def GetValue(self, instance):
        return self.Dataset.GetFeatureValue(self.Feature, instance)

    def GetValueRepresentation(self):
        if self.Dataset.IsNominalFeature(self.Feature):
            return self.Feature[1][self.Value]
        else:
            return self.Value

    def __repr__(self):
        return f"{self.Feature[0]} ? {self.GetValueRepresentation()}"


class EqualThanItem(SingleValueItem):

    def IsMatch(self, instance):
        value = self.GetValue(instance)
        if math.isnan(value):
            return False
        return value == self.Value

    def CompareTo(self, other):
        if isinstance(other, EqualThanItem):
            if self.Value == other.Value:
                return SubsetRelation.Equal
            else:
                return SubsetRelation.Unrelated

        if isinstance(other, DifferentThanItem):
            if self.Value == other.Value:
                return SubsetRelation.Unrelated
            if self.Dataset.IsNominalFeature(self.Feature):
                numberOfValues = len(self.Feature[1])
                if self.Value != other.Value:
                    if numberOfValues == 2:
                        return SubsetRelation.Equal
                    else:
                        return SubsetRelation.Subset

        return SubsetRelation.Unrelated

    def __repr__(self):
        return f"{self.Feature[0]} = {self.GetValueRepresentation()}"


class DifferentThanItem(SingleValueItem):

    def IsMatch(self, instance):
        value = self.GetValue(instance)
        if math.isnan(value):
            return False
        return self.GetValue(instance) != self.Value

    def CompareTo(self, other):
        if isinstance(other, DifferentThanItem):
            if self.Value == other.Value:
                return SubsetRelation.Equal
            else:
                return SubsetRelation.Unrelated

        if isinstance(other, EqualThanItem):
            if self.Value == other.Value:
                return SubsetRelation.Unrelated
            if self.Dataset.IsNominalFeature(self.Feature):
                numberOfValues = len(self.Feature[1])
                if self.Value != other.Value:
                    if numberOfValues == 2:
                        return SubsetRelation.Equal
                    else:
                        return SubsetRelation.Superset

        return SubsetRelation.Unrelated

    def __repr__(self):
        return f"{self.Feature[0]} != {self.GetValueRepresentation()}"


class LessOrEqualThanItem(SingleValueItem):

    def IsMatch(self, instance):
        value = self.GetValue(instance)
        if math.isnan(value):
            return False
        return value <= self.Value

    def CompareTo(self, other):
        if isinstance(other, LessOrEqualThanItem):
            if self.Value == other.Value:
                return SubsetRelation.Equal
            if self.Value > other.Value:
                return SubsetRelation.Superset
            else:
                return SubsetRelation.Subset

        return SubsetRelation.Unrelated

    def __repr__(self):
        return f"{self.Feature[0]} <= {self.GetValueRepresentation():.3f}"


class GreatherThanItem(SingleValueItem):

    def IsMatch(self, instance):
        value = self.GetValue(instance)
        if math.isnan(value):
            return False
        return self.GetValue(instance) > self.Value

    def CompareTo(self, other):
        if isinstance(other, GreatherThanItem):
            if self.Value == other.Value:
                return SubsetRelation.Equal
            if self.Value > other.Value:
                return SubsetRelation.Subset
            else:
                return SubsetRelation.Superset

        return SubsetRelation.Unrelated

    def __repr__(self):
        return f"{self.Feature[0]} > {self.GetValueRepresentation():.3f}"


class ItemBuilder(object):
    def GetItem(self, generalSelector, index):
        return None


class CutPointBasedBuilder(ItemBuilder):
    def GetItem(self, generalSelector, index):
        if not isinstance(generalSelector, CutPointSelector):
            raise Exception(
                f"Unexpected type of selector {generalSelector.__class__.__name__}. Was expecting CutPointSelector")

        if index == 0:
            return LessOrEqualThanItem(generalSelector.Dataset, generalSelector.Feature, generalSelector.CutPoint)
        elif index == 1:
            return GreatherThanItem(generalSelector.Dataset, generalSelector.Feature, generalSelector.CutPoint)
        else:
            raise Exception("Invalid index value for CutPointSelector")


class ValueAndComplementBasedBuilder(ItemBuilder):
    def GetItem(self, generalSelector, index):
        if not isinstance(generalSelector, ValueAndComplementSelector):
            raise Exception(
                f"Unexpected type of selector {generalSelector.__class__.__name__}. Was expecting ValueAndComplementSelector")
        if index == 0:
            return EqualThanItem(generalSelector.Dataset, generalSelector.Feature, generalSelector.Value)
        elif index == 1:
            return DifferentThanItem(generalSelector.Dataset, generalSelector.Feature, generalSelector.Value)
        else:
            raise Exception(
                "Invalid index value for ValueAndComplementSelector")


class MultipleValuesBasedBuilder(ItemBuilder):
    def GetItem(self, generalSelector, index):
        if not isinstance(generalSelector, MultipleValuesSelector):
            raise Exception(
                f"Unexpected type of selector {generalSelector.__class__.__name__}. Was expecting ValueAndComplementSelector")
        if index < 0 or index >= len(generalSelector.Values):
            raise Exception(
                "Invalid index value for MultipleValuesSelector")
        return EqualThanItem(generalSelector.Dataset, generalSelector.Feature, generalSelector.Values[index])


class MultivariateSingleValueItem(Item):
    def __init__(self, dataset, features, value, weights):
        super().__init__(dataset, None)
        self.Value = value
        self.Weights = weights
        self.Features = features
        self.FeaturesHash = sum([hash(feature[0])
                                 for feature in self.Features])
        self._parallel = 0.001


class MultivariateLessOrEqualThanItem(MultivariateSingleValueItem):

    def IsMatch(self, instance):
        instanceValue = self.Dataset.ScalarProjection(
            instance, self.Features, self.Weights)
        if math.isnan(instanceValue):
            return False
        return instanceValue <= self.Value

    def CompareTo(self, other):
        if isinstance(other, MultivariateLessOrEqualThanItem):
            if self.FeaturesHash != other.FeaturesHash:
                return SubsetRelation.Unrelated
            if len(self.Features) != len(other.Features):
                return SubsetRelation.Unrelated

            try:
                proportion = list(other.Weights.values())[
                    0] / list(self.Weights.values())[0]
                for feature in list(self.Weights.keys()):
                    if abs(self.Weights[feature]*proportion - other.Weights[feature]) > self._parallel:
                        return SubsetRelation.Unrelated

                if abs(self.Value * proportion - other.Value) < self._parallel:
                    return SubsetRelation.Equal

                if self.Value * proportion > other.Value:
                    return SubsetRelation.Superset
                else:
                    return SubsetRelation.Subset
            except ZeroDivisionError:
                return SubsetRelation.Unrelated

        return SubsetRelation.Unrelated

    def __repr__(self):
        linearCombination = ' + '.join(
            map(lambda weight: str(self.Weights[weight]) + " * " + weight[0], self.Weights))
        return f"{linearCombination} <= {self.Value}"


class MultivariateGreatherThanItem(MultivariateSingleValueItem):

    def IsMatch(self, instance):
        instanceValue = self.Dataset.ScalarProjection(
            instance, self.Features, self.Weights)
        if math.isnan(instanceValue):
            return False
        return instanceValue <= self.Value

    def CompareTo(self, other):
        if isinstance(other, MultivariateGreatherThanItem):
            if self.FeaturesHash != other.FeaturesHash:
                return SubsetRelation.Unrelated
            if len(self.Features) != len(other.Features):
                return SubsetRelation.Unrelated

            try:
                proportion = list(other.Weights.values())[
                    0] / list(self.Weights.values())[0]
                for feature in list(self.Weights.keys()):
                    if abs(self.Weights[feature]*proportion - other.Weights[feature]) > self._parallel:
                        return SubsetRelation.Unrelated

                if abs(self.Value * proportion - other.Value) < self._parallel:
                    return SubsetRelation.Equal

                if self.Value * proportion > other.Value:
                    return SubsetRelation.Subset
                else:
                    return SubsetRelation.Superset
            except ZeroDivisionError:
                return SubsetRelation.Unrelated

        return SubsetRelation.Unrelated

    def __repr__(self):
        linearCombination = ' + '.join(
            map(lambda weight: str(self.Weights[weight]) + " * " + weight[0], self.Weights))

        return f"{linearCombination} <= {self.Value}"


class MultivariateCutPointBasedBuilder(ItemBuilder):
    def GetItem(self, generalSelector, index):
        if not isinstance(generalSelector, MultivariateCutPointSelector):
            raise Exception(
                f"Unexpected type of selector {generalSelector.__class__.__name__}. Was expecting CutPointSelector")
        if index == 0:
            return MultivariateLessOrEqualThanItem(
                generalSelector.Dataset, generalSelector.Features, generalSelector.CutPoint, generalSelector.Weights)
        elif index == 1:
            return MultivariateGreatherThanItem(
                generalSelector.Dataset, generalSelector.Features, generalSelector.CutPoint, generalSelector.Weights)
        else:
            raise Exception("Invalid index value for CutPointSelector")




class ItemComparer(object):
    def Compare(self, left, right):
        if left.Feature == right.Feature or (not left.Feature and not right.Feature):
            return left.CompareTo(right)
        return SubsetRelation.Unrelated



import math
import random
from .DecisionTree import DecisionTree, DecisionTreeNode
from .WinningSplitSelector import WinningSplitSelector
from .SplitIterator import SplitIterator
from .Helpers import CreateMembershipTupleList, FindDistribution, combine_instances
from .SplitIterator import SplitIteratorProvider, MultivariateSplitIteratorProvider
from .ForwardFeatureIterator import ForwardFeatureIterator
from .DistributionTester import PureNodeStopCondition, AlwaysTrue



class DecisionTreeBuilder():
    def __init__(self, dataset, X, y):
        self.__MinimalInstanceMembership = 0.05
        self.__MinimalSplitGain = 1e-30
        self.__MinimalObjByLeaf = 2
        self.__MaxDepth = -1
        self.__PruneResult = False
        self.__Dataset = dataset
        self.idk = X
        self.__trainInstances = combine_instances(X, y)
        self.__FeatureCount = 0
        self.__StopCondition = PureNodeStopCondition
        self.__distributionEvaluator = None
        self.__OnSelectingFeaturesToConsider = None
        self.__SplitIteratorProvider = SplitIteratorProvider(self.Dataset)
    
    @property
    def MinimalInstanceMembership(self):
        return self.__MinimalInstanceMembership
    
    @property
    def StopCondition(self):
        return self.__StopCondition
    
    @property
    def MaxDepth(self):
        return self.__MaxDepth
    
    @property
    def MinimalObjByLeaf(self):
        return self.__MinimalObjByLeaf
    
    @property
    def SplitIteratorProvider(self):
        return self.__SplitIteratorProvider
    
    @property
    def MinimalSplitGain(self):
        return self.__MinimalSplitGain
    @MinimalSplitGain.setter
    def MinimalSplitGain(self, new_minimal_split_gain):
        self.__MinimalSplitGain = new_minimal_split_gain

    @property
    def Dataset(self):
        return self.__Dataset
    @Dataset.setter
    def Dataset(self, new_dataset):
        self.__Dataset = new_dataset

    @property
    def trainInstances(self):
        return self.__trainInstances
    @trainInstances.setter
    def trainInstances(self, new_train_instances):
        self.__trainInstances = new_train_instances
    
    @property
    def distributionEvaluator(self):
        return self._distributionEvaluator

    @distributionEvaluator.setter
    def distributionEvaluator(self, new_distributionEvaluator):
        self._distributionEvaluator = new_distributionEvaluator

    def Build(self):
        if self.MinimalSplitGain <= 0:
            raise Exception(f"MinimalSplitGain err in Build UniVariate")
            self.MinimalSplitGain = 1e-30

        currentContext = []
        objectMebership = CreateMembershipTupleList(self.trainInstances)
        classFeature = self.Dataset.Class
        result = DecisionTree(self.Dataset)

        filteredObjMembership = list(
            filter(lambda x: x[1] >= self.MinimalInstanceMembership, objectMebership))

        parentDistribution = FindDistribution(
            filteredObjMembership, self.Dataset.Model, self.Dataset.Class)

        result.TreeRootNode = DecisionTreeNode(parentDistribution)
        self.__FillNode(result.TreeRootNode,
                      filteredObjMembership, 0, currentContext)
        return result

    def __FillNode(self, node, instanceTuples, level, currentContext):
        if self.StopCondition(node.Data, self.Dataset.Model, self.Dataset.Class):
            return
        if self.MaxDepth >= 0 and (level >= self.MaxDepth - 1):
            return
        if sum(node.Data) <= self.MinimalObjByLeaf:
            return
        
        whichBetterToFind = 1
        winningSplitSelector = WinningSplitSelector(whichBetterToFind)
        
        sampleFeatures = self.OnSelectingFeaturesToConsider(
            list(map(lambda attribute: attribute[0], self.Dataset.Attributes)), self.FeatureCount)

        for feature in sampleFeatures:
            if feature != self.Dataset.Class[0]:
                splitIterator = self.SplitIteratorProvider.GetSplitIterator(feature)
                splitIterator.Initialize(instanceTuples)
                while splitIterator.FindNext():
                    currentGain = self._distributionEvaluator(
                        node.Data, splitIterator.CurrentDistribution)
                    if currentGain >= self.MinimalSplitGain:
                        winningSplitSelector.EvaluateThis(
                            currentGain, splitIterator, level)
        if winningSplitSelector.IsWinner():
            maxSelector = winningSplitSelector.WinningSelector
            node.ChildSelector = maxSelector
            node.Children = list()
            instancesPerChildNode = CreateChildrenInstances(
                instanceTuples, maxSelector, self.MinimalInstanceMembership)

            for index in range(maxSelector.ChildrenCount):
                childNode = DecisionTreeNode(winningSplitSelector.WinningDistribution[index])
                childNode.Parent = node
                node.Children.append(childNode)

                self.__FillNode(
                    childNode, instancesPerChildNode[index], level + 1, currentContext)

        return


def CreateChildrenInstances(instances, selector, threshold):

    result = list()
    for child in range(selector.ChildrenCount):
        result.append(list(tuple()))

    for instance in instances:
        selection = selector.Select(instance[0])
        if selection is not None:
            for index in range(len(selection)):
                if selection[index] > 0:
                    newMembership = selection[index] * instance[1]
                    if newMembership >= threshold:
                        result[index].append(
                            tuple((instance[0], newMembership)))

    return result



class SelectorContext():
    def __init__(self):
        self.Index = 0
        self.Selector = None


class MultivariateDecisionTreeBuilder(DecisionTreeBuilder):
    def __init__(self, dataset, X, y):
        super().__init__(dataset)
        self.MinimalForwardGain = 0
        self.__trainInstances = combine_instances(X, y)
        self.WMin = 0  # Minimal absolute value for each weight after normalizing
        self.SplitIteratorProvider = MultivariateSplitIteratorProvider(
            self.Dataset)

    def Build(self):
        if self.MinimalSplitGain <= 0:
            self.MinimalSplitGain = 1e-30

        currentContext = []

        objectMebership = CreateMembershipTupleList(self.trainInstances)

        classFeature = self.Dataset.Class

        result = DecisionTree(self.Dataset)

        filteredObjMembership = list(
            filter(lambda x: x[1] >= self.MinimalInstanceMembership, objectMebership))
        
        parentDistribution = FindDistribution(
            filteredObjMembership, self.Dataset.Model, self.Dataset.Class)
        print(f"ParentDist: {parentDistribution}")

        result.TreeRootNode = DecisionTreeNode(parentDistribution)

        self.__FillNode(result.TreeRootNode,
                      filteredObjMembership, 0, currentContext)

        return result

    def FillNode(self, node, instanceTuples, level, currentContext):
        if self.StopCondition(node.Data, self.Dataset.Model, self.Dataset.Class):
            return
        if self.MaxDepth >= 0 and level >= self.MaxDepth - 1:
            return
        if sum(node.Data) <= self.MinimalObjByLeaf:
            return

        whichBetterToFind = 1
        winningSplitSelector = WinningSplitSelector(whichBetterToFind)
        currentGain = 0
        sampleFeatures = self.OnSelectingFeaturesToConsider(
            list(map(lambda attribute: attribute[0], self.Dataset.Attributes)), self.FeatureCount)

        bestFeature = None

        for feature in sampleFeatures:
            splitIterator = self.SplitIteratorProvider.GetSplitIterator(
                feature)
            if not splitIterator:
                raise Exception(f"Undefined iterator for feature {feature}")
            splitIterator.Initialize(instanceTuples)
            while splitIterator.FindNext():
                currentGain = self._distributionEvaluator(node.Data, splitIterator.CurrentDistribution)
                if currentGain >= self.MinimalSplitGain:
                    if winningSplitSelector.EvaluateThis(
                            currentGain, splitIterator, level):
                        bestFeature = self.Dataset.GetAttribute(feature)


        if bestFeature is not None and not self.Dataset.IsNominalFeature(bestFeature):
            sampleFeatures = list(filter(lambda feature: not self.Dataset.IsNominalFeature(
                feature), [self.Dataset.GetAttribute(feature) for feature in sampleFeatures]))
            featureIterator = ForwardFeatureIterator(
                self.Dataset, sampleFeatures)
            featureIterator.Add(bestFeature)
            while featureIterator.FeaturesRemain:
                bestFeature = None
                for features in featureIterator.GetFeatures():
                    candidateFeature = features[0]

                    splitIterator = self.SplitIteratorProvider.GetMultivariateSplitIterator(
                        features, self.WMin)
                    if not splitIterator:
                        raise Exception(
                            f"Undefined iterator for features {','.join(map(lambda feature: feature[0],features))}")

                    valid = splitIterator.InitializeMultivariate(
                        instanceTuples, node)

                    if not valid:
                        break

                    while splitIterator.FindNext():
                        currentGain = self._distributionEvaluator(
                            node.Data, splitIterator.CurrentDistribution)
                        if currentGain >= self.MinimalSplitGain and (currentGain - winningSplitSelector.MinStoredValue) >= self.MinimalForwardGain:
                            if winningSplitSelector.EvaluateThis(currentGain, splitIterator, level):
                                bestFeature = candidateFeature
                if not bestFeature:
                    break
                else:
                    featureIterator.Add(bestFeature)

        if winningSplitSelector.IsWinner():
            maxSelector = winningSplitSelector.WinningSelector
            node.ChildSelector = maxSelector
            node.Children = list()
            instancesPerChildNode = CreateChildrenInstances(
                instanceTuples, maxSelector, self.MinimalInstanceMembership)

            for index in range(maxSelector.ChildrenCount):
                childNode = DecisionTreeNode(winningSplitSelector.WinningDistribution[index])
                childNode.Parent = node
                node.Children.append(childNode)

                self.__FillNode(
                    childNode, instancesPerChildNode[index], level + 1, currentContext)
        

        return



import math
import random
import operator
import numpy as np
import pandas as pd
from itertools import chain

def SumMatrix(matrix):
    if not matrix or len(matrix) == 0:
        return 0

    return sum(map(sum, matrix))


def ArgMin(source):
    if not source:
        return -1
    minValue = min(source)
    return source.index(minValue)


def ArgMax(source):
    if not source:
        return -1
    maxValue = max(source)
    return source.index(maxValue)


def MultiplyBy(array, value):
    return list(map(lambda element: element * value, array))


def AddTo(a, b):
    if len(a) != len(b):
        raise Exception
    return list(map(operator.add, a, b))


def Substract(a, b):
    if len(a) != len(b):
        raise Exception
    return list(map(operator.sub, a, b))


def CreateMembershipTupleList(instances):
    tupleList =  [(x,1.0) for x in instances]
    return tupleList

def combine_instances(X, y):
    combined_list = []
    for i,val in enumerate(X):
        combined_list.append(__chain_together(val, y[i]))
    result = np.asarray(combined_list, dtype= np.object)
    return result

def __chain_together(a, b):
    return list(chain(*[a,b]))

#Find Distribution of class values in dataset, i.e how many positive vs negative instances there are
def FindDistribution(source, model, classFeature):
    if isinstance(classFeature[1], str):
        raise Exception("Cannot find distribution for non-nominal class")

    result = [0]*len(classFeature[1])
    classIdx = model.index(classFeature)

    for element in source:
        value = classFeature[1].index(element[0][classIdx])
        
        if value >= 0:
            result[value] += element[1]
        else:
            continue

    return result

def convert_to_ndarray(y):
    new_y = []
    for y_elem in y:
         new_y.append(np.array([y_elem], dtype=np.object))
    pd_series = pd.Series(new_y, name='class')
    return pd_series

def get_col_dist(source):
    elems = set()
    for elem in source:
        elems.add(elem)    
    return sorted(elems)

def get_idx_val(source, instance):
    for idx, val in enumerate(source):
        if instance == val:
            return idx
    raise Exception(f"Nominal value inside training dataset not found inside testing dataset")




from .DecisionTree import DecisionTree, DecisionTreeNode
from .Helpers import MultiplyBy, AddTo


class DecisionTreeClassifier(object):
    def __init__(self, tree):
        self.__DecisionTree = tree
        self.__Model = tree.Model

    @property
    def DecisionTree(self):
        return self.__DecisionTree
    @DecisionTree.setter
    def DecisionTree(self, new_decision_tree):
        self.__DecisionTree = new_decision_tree

    @property
    def Model(self):
        return self.__Model
    @Model.setter
    def Model(self, new_model):
        self.__Model = new_model

    def __Classify(self, instance):
        classification = self.ClassifyInstance(
            self.DecisionTree.TreeRootNode, instance, 1)
        return MultiplyBy(classification, (1/sum(classification)))

    def __ClassifyInstance(self, node, instance, instanceMembership):
        if node.IsLeaf:
            return MultiplyBy(node.Data, instanceMembership)

        childrenSelection = node.ChildSelector.Select(instance)
        result = None
        if (childrenSelection != None):
            if (len(childrenSelection) != len(node.Children)):
                raise Exception("Child index is out of range")

            for i in range(len(childrenSelection)):
                selection = childrenSelection[i]
                if selection > 0:
                    child = node.Children[i]
                    childValue = __ClassifyInstance(
                        child, instance, instanceMembership)
                    if result != None:
                        result = AddTo(result, childValue)
                    else:
                        result = childValue
        else:
            totalNodeMembership = sum(node.Data)
            for i in range(len(node.Children)):
                child = node.Children[i]
                childMembership = sum(node.Children[i].Data)
                childValue = __ClassifyInstance(
                    child, instance, childMembership / (totalNodeMembership * instanceMembership))
                if result != None:
                    result = AddTo(result, childValue)
                else:
                    result = childValue
                    
        return result



from copy import copy
from .Item import SubsetRelation
from .FilteredCollection import FilteredCollection
from .DecisionTreeBuilder import SelectorContext
from .Item import CutPointBasedBuilder, MultipleValuesBasedBuilder, ValueAndComplementBasedBuilder, MultivariateCutPointBasedBuilder
from .FeatureSelectors import CutPointSelector, MultipleValuesSelector, ValueAndComplementSelector, MultivariateCutPointSelector
from itertools import chain
from collections import OrderedDict


class EmergingPattern(object):
    def __init__(self, dataset, items=None):
        self.__Dataset = dataset
        self.__Model = self.Dataset.Model
        self.__Items = None
        if not items:
            self.Items = list()
        else:
            self.Items = items
        self.__Counts = []
        self.__Supports = []
    
    @property
    def Dataset(self):
        return self.__Dataset
    @Dataset.setter
    def Dataset(self, new_dataset):
        self.__Dataset = new_dataset
    
    @property
    def Model(self):
        return self.__Model
    @Model.setter
    def Model(self, new_model):
        self.__Model = new_model

    @property
    def Items(self):
        return self.__Items
    @Items.setter
    def Items(self, new_items):
        self.__Items = new_items
    
    @property
    def Counts(self):
        return self.__Counts
    @Counts.setter
    def Counts(self, new_counts):
        self.__Counts = new_counts
    
    @property
    def Supports(self):
        return self.__Supports
    @Supports.setter
    def Supports(self, new_supports):
        self.__Supports = new_supports

    def IsMatch(self, instance):
        for item in self.Items:
            if not item.IsMatch(instance):
                return False
        return True

    def __UpdateCountsAndSupport(self, instances):
        matchesCount = [0]*len(self.Dataset.Class[1])

        for instance in instances:
            if(self.IsMatch(instance)):
                matchesCount[instance[self.Dataset.GetClassIdx()]] += 1
        self.Counts = matchesCount
        self.Supports = self.CalculateSupports(matchesCount)

    def CalculateSupports(self, data, classFeatureParam=None):
        if classFeatureParam == None:
            classInfo = self.Dataset.ClassInformation
            result = copy(data)
            for i in range(len(result)):
                if classInfo.Distribution[i] != 0:
                    result[i] /= classInfo.Distribution[i]
                else:
                    result[i] = 0
            return result
        else:
            classFeature = self.Dataset.ClassInformation.Feature
            featureInformation = self.Dataset.ClassInformation.Distribution
            result = copy(data)
            for i in range(len(result)):
                if featureInformation[i] != 0:
                    result[i] /= featureInformation[i]
                else:
                    result[i] = 0
            return result
                

    def __Clone(self):

        result = EmergingPattern(self.Dataset, self.Items)
        result.Counts = copy(self.Counts)
        result.Supports = copy(self.Supports)
        return result

    def __repr__(self):
        return self.BaseRepresentation() + "\n" + self.SupportInfo()

    def BaseRepresentation(self):
        return ' AND '.join(map(lambda item: item.__repr__(), self.Items))

    def SupportInfo(self):
        return ' '.join(map(lambda count, support, className: f"{className} count: {str(count)} support: [{str(round(support,2))}] ", self.Counts, self.Supports, self.Dataset.Class[1]))

    def ToString(self):
        dictOfPatterns = {"Pattern": self.BaseRepresentation()}

        dictOfClasses = {self.Dataset.Class[1][i]+" Count": self.Counts[i]
                         for i in range(0, len(self.Dataset.Class[1]))}

        dictOfClasses.update({self.Dataset.Class[1][i]+" Support": self.Supports[i]
                              for i in range(0, len(self.Dataset.Class[1]))})

        for key in sorted(dictOfClasses.keys()):
            dictOfPatterns.update({key: dictOfClasses[key]})

        return dictOfPatterns


class EmergingPatternCreator(object):
    def __init__(self, dataset):
        self.__Dataset = dataset
        self.__builderForType = {
            CutPointSelector: CutPointBasedBuilder,
            MultipleValuesSelector: MultipleValuesBasedBuilder,
            ValueAndComplementSelector: ValueAndComplementBasedBuilder,
            MultivariateCutPointSelector: MultivariateCutPointBasedBuilder
        }
    
    @property
    def Dataset(self):
        return self.__Dataset
    @Dataset.setter
    def Dataset(self, new_dataset):
        self.__Dataset = new_dataset

    def __Create(self, contexts):
        pattern = EmergingPattern(self.Dataset)
        for context in contexts:
            childSelector = context.Selector
            builder = self.__builderForType[context.Selector.__class__]()
            item = builder.GetItem(childSelector, context.Index)
            pattern.Items.append(item)
        return pattern

    def ExtractPatterns(self, treeClassifier, patternFound):
        context = list()
        self.__DoExtractPatterns(
            treeClassifier.DecisionTree.TreeRootNode, context, patternFound)

    def __DoExtractPatterns(self, node, contexts, patternFound):
        if node.IsLeaf:
            newPattern = self.__Create(contexts)
            newPattern.Counts = node.Data
            newPattern.Supports = newPattern.CalculateSupports(node.Data)
            if patternFound is not None:
                patternFound(newPattern)
        else:
            for index in range(len(node.Children)):
                selectorContext = SelectorContext()
                selectorContext.Index = index
                selectorContext.Selector = node.ChildSelector
                context = selectorContext

                contexts.append(context)
                self.__DoExtractPatterns(node.Children[index], contexts, patternFound)
                contexts.remove(context)


class EmergingPatternComparer(object):
    def __init__(self, itemComparer):
        self.__Comparer = itemComparer
    @property
    def Comparer(self):
        return self.__Comparer
    @Comparer.setter
    def Comparer(self, new_comparer):
        self.__Comparer = new_comparer

    def Compare(self, leftPattern, rightPattern):
        directSubset = self.IsSubset(leftPattern, rightPattern)
        inverseSubset = self.IsSubset(rightPattern, leftPattern)
        if (directSubset and inverseSubset):
            return SubsetRelation.Equal
        elif (directSubset):
            return SubsetRelation.Subset
        elif (inverseSubset):
            return SubsetRelation.Superset
        else:
            return SubsetRelation.Unrelated

    def IsSubset(self, pat1, pat2):
        def f(x, y):
            relation = self.Comparer.Compare(self, x, y)
            return relation == SubsetRelation.Equal or relation == SubsetRelation.Subset

        for x in pat2.Items:
            all_bool = False
            for y in pat1.Items:
                if f(y, x):
                    all_bool = True
                    break
            if not all_bool:
                return False
        return True

class EmergingPatternSimplifier(object):
    def __init__(self, itemComparer):
        self.__comparer = itemComparer
        self.__collection = FilteredCollection(
            self.__comparer, SubsetRelation.Subset)

    def Simplify(self, pattern):
        resultPattern = EmergingPattern(pattern.Dataset)
        resultPattern.Counts = copy(pattern.Counts)
        resultPattern.Supports = copy(pattern.Supports)
        self.__collection.current = resultPattern.Items
        self.__collection.AddRange(pattern.Items)
        return resultPattern



from abc import ABC, abstractmethod 
import os
import random
import math
import pandas as pd
import sys
import numpy as np
import pandas as pd
from .Helpers import get_col_dist


class Dataset(ABC):
    @abstractmethod
    def GetAttribute(self, attribute):
        pass

class PandasDataset(Dataset):
    def __init__(self, X,y):
        self.Model = self.get_model_list(X,y)
        self.Instances = self.combine_X_y(X,y)

    def GetAttribute(self, attribute):
        pass
    def get_feature_info(self, feature):
        if self.get_feature_col_type(feature) == 'Nominal':
            return get_col_dist(feature)
        else:
            return self.get_feature_col_type(feature)
    def get_feature_col_type(self, feature):
        if isinstance(feature[0], np.int32) or isinstance(feature[0], np.int64):
            return 'integer'
        elif isinstance(feature[0], np.float32) or isinstance(feature[0], np.float64):
            return 'real'
        elif isinstance(feature[0], str):
            return 'Nominal'
        else:
            raise Exception(f"Unsupported data type in feature {feature}")

    def get_model_list(self, X, y):
        result = [(feat_name, self.get_feature_info(X[f'{feat_name}'])) for feat_name in X]
        class_res = [(feat_name, self.get_feature_info(y[f'{feat_name}'])) for feat_name in y]
        result.append(class_res[0])
        return result
    
    def combine_X_y(self, X, y):
        y_name = list(y.columns)[0] 
        instances_df = X.copy(deep=True)
        instances_df[f'{y_name}'] = y[f'{y_name}']
        return instances_df.values

    @property
    def Attributes(self):
        return list(
            filter(lambda attr: attr[0].strip().lower() != 'class', self.Model))

    @property
    def Class(self):
        return list(
            filter(lambda attr: attr[0].strip().lower() == 'class', self.Model))[0]

    @property
    def AttributesInformation(self):
        return list(
            map(lambda attr: FeatureInformation(self, attr), self.Attributes))

    @property
    def ClassInformation(self):
        return FeatureInformation(self, self.Class)

    def GetAttribute(self, attribute):
        return list(filter(lambda attr: attr[0] == attribute, self.Model))[0]

    def GetAttributeNames(self):
        return list(
            map(lambda attribute: attribute[0], self.Model))

    def GetNominalValues(self, feature):
        attribute = list(
            filter(lambda attr: attr[0].lower() == feature.lower(), self.Model))

        if isinstance(attribute[0][1], list):
            return attribute[0][1]
        else:
            return None

    def GetValueOfIndex(self, feature, index):
        values = self.GetNominalValues(feature)
        if not values:
            return None
        else:
            return values[index]

    def GetIndexOfValue(self, feature, value):
        values = self.GetNominalValues(feature)
        if not values:
            return -1
        else:
            return values.index(value)

    def GetClasses(self):
        return self.GetNominalValues('class')

    def GetClassIdx(self):
        return self.GetFeatureIdx(self.Class)

    def GetFeatureIdx(self, feature):
        return self.Model.index(feature)

    def IsNominalFeature(self, feature):
        if isinstance(feature[1], str) and feature[1].lower() == 'string':
            raise Exception("String attributes are not supported!")
        return isinstance(feature[1], list)

    def GetFeatureValue(self, feature, instance):
        if isinstance(feature[1], list):
            return self.GetIndexOfValue(feature[0], instance[self.GetFeatureIdx(feature)])
        elif feature[1].lower() in ['numeric', 'real', 'integer']:
            return instance[self.GetFeatureIdx(feature)]
        else:
            raise Exception(
                "Attribute must be either nominal, numeric, real or integer")
    
    def GetClassValue(self, y_instance):
        return self.GetIndexOfValue(self.Class[0], y_instance[0])

    def IsMissing(self, feature, instance):
        value = self.GetFeatureValue(feature, instance)

        if self.IsNominalFeature(feature):
            return value < 0
        else:
            return (not value or value == math.nan)   
            

class FileDataset(Dataset):
    from .FileManipulation import ReadARFF, ReadDAT, GetFromFile, get_dataframe_from_arff

    def __init__(self, file):
        arffFile = GetFromFile(file)

        attributes = arffFile['attributes']
        if not list(filter(lambda attr: attr[0].strip().lower() == 'class', attributes)):
            raise Exception(
                f"This dataset does not contain an explicit class attribute!")

        self.Relation = arffFile['relation']

        self.Model = arffFile['attributes']

        self.__Instances = get_dataframe_from_arff(arffFile).values
        self.Features = None
    
    @property
    def Instances(self):
        return self.__Instances

    @property
    def Attributes(self):
        return list(
            filter(lambda attr: attr[0].strip().lower() != 'class', self.Model))

    @property
    def Class(self):
        return list(
            filter(lambda attr: attr[0].strip().lower() == 'class', self.Model))[0]

    @property
    def AttributesInformation(self):
        return list(
            map(lambda attr: FeatureInformation(self, attr), self.Attributes))

    @property
    def ClassInformation(self):
        return FeatureInformation(self, self.Class)

    def GetAttribute(self, attribute):
        return list(filter(lambda attr: attr[0] == attribute, self.Model))[0]

    def GetAttributeNames(self):
        return list(
            map(lambda attribute: attribute[0], self.Model))

    def GetNominalValues(self, feature):
        attribute = list(
            filter(lambda attr: attr[0].lower() == feature.lower(), self.Model))

        if isinstance(attribute[0][1], list):
            return attribute[0][1]
        else:
            return None

    def GetValueOfIndex(self, feature, index):
        values = self.GetNominalValues(feature)
        if not values:
            return None
        else:
            return values[index]

    def GetIndexOfValue(self, feature, value):
        values = self.GetNominalValues(feature)
        if not values:
            return -1
        else:
            return values.index(value)

    def GetClasses(self):
        return self.GetNominalValues('class')

    def GetClassIdx(self):
        return self.GetFeatureIdx(self.Class)

    def GetFeatureIdx(self, feature):
        return self.Model.index(feature)

    def IsNominalFeature(self, feature):
        if isinstance(feature[1], str) and feature[1].lower() == 'string':
            raise Exception("String attributes are not supported!")
        return isinstance(feature[1], list)

    def GetFeatureValue(self, feature, instance):
        if isinstance(feature[1], list):
            return self.GetIndexOfValue(feature[0], instance[self.GetFeatureIdx(feature)])
        elif feature[1].lower() in ['numeric', 'real', 'integer']:
            return instance[self.GetFeatureIdx(feature)]
        else:
            raise Exception(
                "Attribute must be either nominal, numeric, real or integer")
    
    def GetClassValue(self, y_instance):
        return self.GetIndexOfValue(self.Class[0], y_instance[0])

    def IsMissing(self, feature, instance):
        value = self.GetFeatureValue(feature, instance)

        if self.IsNominalFeature(feature):
            return value < 0
        else:
            return (not value or value == math.nan)

    def ScalarProjection(self, instance, features, weights):

        if len(list(filter(lambda feature: self.IsMissing(feature, instance), features))) > 0:
            return math.nan

        result = sum([weights[feature] * self.GetFeatureValue(feature, instance)
                      for feature in features])

        return result

class FeatureInformation(object):
    def __init__(self, dataset, feature):
        self.__Dataset = dataset
        self.__Feature = feature
        self.__MinValue = 0
        self.__MaxValue = 0
        self.__Distribution = []
        self.__ValueProbability = []
        self.__Ratio = []

        self.Initialize()
    
    @property
    def Dataset(self):
        return self.__Dataset
    @Dataset.setter
    def Dataset(self, new_dataset):
        self.__Dataset = new_dataset

    @property
    def Feature(self):
        return self.__Feature

    @property
    def Distribution(self):
        return self.__Distribution
    @Distribution.setter
    def Distribution(self, new_distribution):
        self.__Distribution = new_distribution

    def Initialize(self):

        featureIdx = self.Dataset.GetFeatureIdx(self.Feature)

        nonMissingValues = list(filter(lambda instance: not self.Dataset.IsMissing(
            self.Feature, instance), self.Dataset.Instances))

        self.MissingValueCount = len(
            self.Dataset.Instances) - len(nonMissingValues)

        if self.Dataset.IsNominalFeature(self.Feature):

            self.Distribution = [0]*len(self.Feature[1])
            for value in range(len(self.Distribution)):
                self.Distribution[value] = len(
                    list(filter(lambda instance: self.Dataset.GetFeatureValue(self.Feature, instance) == value and not self.Dataset.IsMissing(self.Feature, instance), self.Dataset.Instances)))
            self.ValueProbability = list(
                map(lambda value: value / sum(self.Distribution), self.Distribution))
            self.Ratio = list(
                map(lambda value: value / min(self.Distribution), self.Distribution))

        else:
            if len(nonMissingValues) > 0:
                self.MinValue = min(list(map(
                    lambda instance: instance[featureIdx], nonMissingValues)))
                self.MaxValue = max(list(map(
                    lambda instance: instance[featureIdx], nonMissingValues)))
            else:
                self.MinValue = 0
                self.MaxValue = 0
            




from .Item import SubsetRelation

class FilteredCollection(object):
    def __init__(self, comparer, relationToFind, resultCollection=None):
        self.__comparer = comparer
        self.__relationToFind = relationToFind
        self.__inverseRelation = SubsetRelation.Unrelated

        if self.__relationToFind == SubsetRelation.Superset:
            self.__inverseRelation = SubsetRelation.Subset
        elif self.__relationToFind == SubsetRelation.Subset:
            self.__inverseRelation = SubsetRelation.Superset
        elif self.__relationToFind == SubsetRelation.Equal:
            self.__inverseRelation = SubsetRelation.Different

        self.__current = None
        if not resultCollection:
            self.__current = list()
        else:
            self.__current = resultCollection
    
    @property
    def current(self):
        return self.__current
    @current.setter
    def current(self, new_current):
        self.__current = new_current

    def __Add(self, item):
        if (self.__relationToFind != SubsetRelation.Unrelated):
            i = 0
            while i < len(self.__current):

                relation = self.__comparer(item, self.__current[i])
                if (relation == SubsetRelation.Equal or relation == self.__inverseRelation):
                    return
                elif (relation == self.__relationToFind):
                    self.__current.remove(self.__current[i])
                else:
                    i += 1
        
        self.__current.append(item)

    def GetItems(self):
        return self.__current

    def AddRange(self, items):
        for item in items:
            self.__Add(item)


import os
import argparse
import numpy as np
import pandas as pd

from tqdm import tqdm, trange
from core.PBC4cip import PBC4cip
from core.FileManipulation import WritePatternsBinary, WritePatternsCSV, ReadPatternsBinary
from core.FileManipulation import WriteResultsCSV, returnX_y, get_dataframe_from_arff, GetFromFile, convert_dat_to_csv
from core.DecisionTreeBuilder import DecisionTreeBuilder, MultivariateDecisionTreeBuilder
from core.PatternMiner import PatternMinerWithoutFiltering
from core.PatternFilter import MaximalPatternsGlobalFilter
from core.DistributionEvaluator import Hellinger, MultiClassHellinger, QuinlanGain
from core.Evaluation import obtainAUCMulticlass
from core.Helpers import ArgMax, convert_to_ndarray, get_col_dist, get_idx_val
from core.Dataset import Dataset, FileDataset, PandasDataset
from datetime import datetime

def CheckSuffix(file, suffix):
    if not suffix or len(suffix) == 0:
        return True
    if not suffix in file or len(suffix) >= len(file):
        return False
    filename, file_extension = os.path.splitext(file)
    return filename[(len(suffix)*-1):] == suffix


def GetFilesFromDirectory(directory):
    print(directory)
    files = []
    if os.path.isdir(directory):
        for r, d, f in os.walk(directory):
            for file in f:
                files.append(os.path.join(r, file))
        return files
    else:
        raise Exception(f"Directory {directory} is not valid.")

def import_data(trainFile, testFile):
    train = pd.read_csv(trainFile, sep= ',') 
    test = pd.read_csv(testFile, sep= ',')

    return train, test

def split_data(train, test):
    X_train = train.iloc[:,  0:train.shape[1]-1]
    y_train =  train.iloc[:, train.shape[1]-1 : train.shape[1]]

    X_test = test.iloc[:,  0:test.shape[1]-1]
    y_test =  test.iloc[:, test.shape[1]-1 : test.shape[1]]

    return X_train, y_train, X_test, y_test

def score(predicted, y):
        y_class_dist = get_col_dist(y[f'{y.columns[0]}'])
        real = list(map(lambda instance: get_idx_val(y_class_dist, instance), y[f'{y.columns[0]}']))
        numClasses = len(y_class_dist)
        confusion = [[0]*2 for i in range(numClasses)]
        classified_as = 0
        error_count = 0

        for i in range(len(real)):
            if real[i] != predicted[i]:
                error_count = error_count + 1
            confusion[real[i]][predicted[i]] = confusion[real[i]][predicted[i]] + 1

        acc = 100.0 * (len(real) - error_count) / len(real)
        auc = obtainAUCMulticlass(confusion, numClasses)

        return confusion, acc, auc

def score_txtfile(predicted, y, dataset):
    real = list(map(lambda instance: dataset.GetClassValue(instance), y))
    numClasses = len(dataset.Class[1])
    confusion = [[0]*2 for i in range(numClasses)]
    classified_as = 0
    error_count = 0

    for i in range(len(real)):
        if real[i] != predicted[i]:
            error_count = error_count + 1
        confusion[real[i]][predicted[i]] = confusion[real[i]][predicted[i]] + 1

    acc = 100.0 * (len(real) - error_count) / len(real)
    auc = obtainAUCMulticlass(confusion, numClasses)

    return confusion, acc, auc


def show_results(confusion, acc, auc, numPatterns):
    print()
    for i in range(len(confusion[0])):
        for j in range(len(confusion[0])):
            print(f"{confusion[i][j]} ", end='')
        print("")
    print(f"acc: {acc} , auc: {auc} , numPatterns: {numPatterns}")

def Train_and_test(X_train, y_train, X_test, y_test, treeCount, multivariate, filtering, dataset=None):
    classifier = PBC4cip(tree_count=treeCount, filtering = filtering, file_dataset=dataset)
    patterns = classifier.fit(X_train, y_train)

    y_pred = classifier.predict(X_test)
    confusion, acc, auc = classifier.score(y_pred, y_test)

    
    return patterns, confusion, acc, auc

def test_PBC4cip(trainFile, outputDirectory, treeCount, multivariate, filtering, testFile, resultsId, delete): 
    #Uncomment this to work with text files instead of dataframes  
    """
    X_train, y_train = returnX_y(trainFile)
    X_test, y_test = returnX_y(testFile)
    file_dataset = FileDataset(trainFile)
    classifier = PBC4cip(tree_count=treeCount, multivariate=multivariate, filtering=filtering, file_dataset=trainFile)
    patterns = classifier.fit(X_train, y_train)
    y_test_scores = classifier.score_samples(X_test)
    y_pred = classifier.predict(y_test_scores)
    confusion, acc, auc = score_txtfile(y_pred, y_test, FileDataset(trainFile))
    """
    train_df, test_df = import_data(trainFile, testFile)
    X_train, y_train, X_test, y_test = split_data(train_df, test_df)
    classifier = PBC4cip(tree_count=treeCount, multivariate=multivariate, filtering=filtering)
    patterns = classifier.fit(X_train, y_train)

    y_test_scores = classifier.score_samples(X_test)
    

    y_pred = classifier.predict(y_test_scores)
    confusion, acc, auc = score(y_pred, y_test)
    
    WritePatternsCSV(patterns, trainFile, outputDirectory)
    WritePatternsBinary(patterns, trainFile, outputDirectory)
    WriteResultsCSV(confusion, acc, auc, len(patterns), testFile, outputDirectory, resultsId, filtering)
    show_results(confusion, acc, auc, len(patterns))


def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def Execute(args):
    testing_files = []
    training_files = []

    if args.training_files:
        training_files = args.training_files

    if args.training_directory:
        training_files = list(set().union(training_files, list(filter(
            lambda file: CheckSuffix(file, args.training_file_suffix), GetFilesFromDirectory(args.training_directory)))))

    if args.input_files:
        testing_files = args.input_files

    if args.input_directory:
        testing_files = list(set().union(testing_files, list(filter(
            lambda file: CheckSuffix(file, args.test_file_suffix), GetFilesFromDirectory(args.input_directory)))))

    training_files.sort()
    testing_files.sort()

    print(
        f"Training files detected (*{args.training_file_suffix}.[arff|dat]): {len(training_files)}")
    print(
        f"Testing files detected (*{args.test_file_suffix}.[arff|dat]): {len(testing_files)}")

    print("===============================================================================")
    tra = trange(len(training_files), desc='Training and Testing Files...', leave=True, unit="dataset")

    now = datetime.now()
    resultsId = now.strftime("%Y%m%d%H%M%S")

    for f in tra:
        tra.set_description(f"Working from {training_files[f]}")
        test_PBC4cip(training_files[f], args.output_directory, args.tree_count, args.multivariate,
            args.filtering,  testing_files[f], resultsId, args.delete_binary )


if __name__ == '__main__':

    defaultDataDir = os.path.join(os.path.normpath(
        os.path.join(os.getcwd(), os.pardir)), "data", "python")
    defaultOutputDir = os.path.join(os.path.normpath(
        os.path.join(os.getcwd(), os.pardir)), "output")

    defaultTrainingFiles = list()
    defaultTestingFiles = list()
    parser = argparse.ArgumentParser(
        description="Process class imbalanced datasets using PBC4cip.")

    parser.add_argument("--training-files",
                        type=str,
                        metavar="<*.dat/*.arff>",
                        nargs="+",
                        help="a file or files that are going to be used to train the classifier")

    parser.add_argument("--training-directory",
                        type=str,
                        metavar="'"+defaultDataDir+"'",
                        help="the directory with files to be used to train the classifier")

    parser.add_argument("--input-files",
                        type=str,
                        metavar="<*.dat/*.arff>",
                        nargs="+",
                        help="a file or files to be classified")

    parser.add_argument("--input-directory",
                        type=str,
                        metavar="'"+defaultDataDir+"'",
                        help="the directory with files to be classified")

    parser.add_argument("--output-directory",
                        type=str,
                        metavar="'"+defaultOutputDir+"'",
                        default=defaultOutputDir,
                        help="the output directory for the patterns")

    parser.add_argument("--multivariate",
                        type=str2bool,
                        const=True,
                        default=False,
                        nargs='?',
                        help="states if multivariate tree builder variant is to be used")

    parser.add_argument("--delete-binary",
                        type=str2bool,
                        default=True,
                        nargs='?',
                        help="states if binary file is to be deleted after execution")

    parser.add_argument("--tree-count",
                        type=int,
                        metavar="n",
                        default=100,
                        help="indicates the number of trees that will be used")

    parser.add_argument("--filtering",
                        type = str2bool,
                        const=True,
                        default=False,
                        nargs='?',
                        help="Decides wether the found patterns are to be filtered or not")

    parser.add_argument("--test-file-suffix",
                        type=str,
                        metavar="'tst'",
                        default="tst",
                        help="states which suffix will indicate the test files")

    parser.add_argument("--training-file-suffix",
                        type=str,
                        metavar="'tra'",
                        default="tra",
                        help="states which suffix will indicate the training files")

    

    args = parser.parse_args()

    print("==========================================================")
    print("         o--o  o--o    o-o o  o                           ")
    print("         |   | |   |  /    |  |      o                    ")
    print("         O--o  O--o  O     o--O  o-o   o-o                ")
    print("         |     |   |  \       | |    | |  |               ")
    print("         o     o--o    o-o    o  o-o | O-o                ")
    print("                                       |                  ")
    print("                                       o                  ")
    print("==========================================================")

    if not args.training_files and not args.training_directory and not args.input_files and not args.input_directory:
        parser.print_help()
    else:
        Execute(args)



import os
import pickle
import csv
from tqdm import tqdm
from io import StringIO, BytesIO
import pandas as pd


def ReadARFF(file):
    import arff
    return arff.loads(open(file))


def ReadAndDeleteARFF(file):
    dataset = ReadARFF(file)
    if os.path.exists(file):
        os.remove(file)
    else:
        print(f"Can not delete the file '{file}' as it doesn't exists")
    return dataset


def ReadDAT(file):
    f = open(file).readlines()
    new_f = str(file+'.arff')
    with open(new_f, 'w') as new_file:
        for line in f:
            if line != '\n':
                newline = line
                if newline.split()[0] == '@inputs' or newline.split()[0] == '@output':
                    continue
                elif newline.split()[0] == '@attribute':
                    newline = newline.replace(
                        'real', 'real ').replace(
                        'REAL', 'REAL ').replace('  ', ' ')
                    if newline.split()[2].lower() in ['numeric', 'real', 'integer', 'string']:
                        newline = newline.split(
                        )[0]+' '+newline.split()[1]+' '+newline.split()[2]+'\n'
                new_file.write(newline)
    return ReadAndDeleteARFF(new_f)

def convert_dat_to_csv(file):
    f = open(file).readlines()
    new_f = str(file[0:len(file)-4]+'.csv')
    column_names = []
    with open(new_f, 'w') as new_file:
        for line in f:
            if line != '\n':
                newline = line
                if newline.split()[0] == '@outputs' or newline.split()[0] == '@output' or newline.split()[0] == '@attribute' \
                or newline.split()[0] == '@relation' or newline.split()[0] == '@data' or newline.split()[0] == '@relation':
                    continue
                elif newline.split()[0] == '@inputs' or newline.split()[0] == '@input':
                    newline = newline.replace(
                        '@inputs', '@input'
                    ).replace('@input', '').replace(' ', '').replace('\n','')
                    newline = newline + ',Class\n'
                else :
                    newline.replace(' ', '')
                new_file.write(newline)
    return None


def ReadPatternsBinary(originalFile, outputDirectory, delete, suffix=None):
    patterns = list()

    if not suffix:
        suffix = ""

    name = os.path.splitext(os.path.basename(originalFile))[0]
    name = os.path.join(
        outputDirectory, name[:len(name)-len(suffix)]+'.pypatterns')

    if os.path.exists(name):
        input_file = open(name, "rb")
        patternCount = pickle.load(input_file)
        for pattern in tqdm(range(patternCount), desc=f"Reading patterns from {name}", unit="pat", leave=False):
            try:
                pattern_in = pickle.load(input_file)
                patterns.append(pattern_in)
            except EOFError:
                break
    else:
        raise Exception(
            f"File '{name}'' not found! Please extract patterns first!")
    return patterns


def WritePatternsBinary(patterns, originalFile, outputDirectory, suffix=None):
    if not patterns or len(patterns) == 0:
        return ""
    if not suffix:
        suffix = ""
    if not os.path.exists(outputDirectory):
        print(f"Creating output directory: {outputDirectory}")
        os.makedirs(outputDirectory)

    name = os.path.splitext(os.path.basename(originalFile))[0]
    name = os.path.join(
        outputDirectory, name[:len(name)-len(suffix)]+'.pypatterns')

    action = "Writing"
    if os.path.exists(name):
        action = "Overwriting"
        os.remove(name)

    patterns_out = open(name, "wb")
    pickle.dump(len(patterns), patterns_out)
    for pattern in tqdm(patterns, desc=f"{action} patterns to {name}...", unit="pattern", leave=False):
        pickle.dump(pattern, patterns_out)
        patterns_out.flush()
    patterns_out.close()

    return name


def WritePatternsCSV(patterns, originalFile, outputDirectory, suffix=None):
    if not patterns or len(patterns) == 0:
        return ""
    if not suffix:
        suffix = ""
    if not os.path.exists(outputDirectory):
        print(f"Creating output directory: {outputDirectory}")
        os.makedirs(outputDirectory)

    name = os.path.splitext(os.path.basename(originalFile))[0]
    name = os.path.join(outputDirectory, name[:len(name)-len(suffix)]+'.csv')

    action = "Writing"
    if os.path.exists(name):
        action = "Overwriting"
        os.remove(name)

    patterns_out = open(name, "w", newline='\n', encoding='utf-8')
    fields = list(patterns[0].ToString().keys())
    pattern_writer = csv.DictWriter(patterns_out, fieldnames=fields)
    pattern_writer.writeheader()
    for pattern in tqdm(patterns, desc=f"{action} patterns to {name}...", unit="pattern", leave=False):
        pattern_writer.writerow(pattern.ToString())

    patterns_out.close()

    return name

def WriteResultsCSV(confusion, acc, auc, numPatterns, originalFile, outputDirectory, resultsId, filtering):
    if not os.path.exists(outputDirectory):
        print(f"Creating output directory: {outputDirectory}")
        os.makedirs(outputDirectory)

    datasetName = os.path.splitext(os.path.basename(originalFile))[0]
    name = os.path.join(outputDirectory, f"TestsResults{resultsId}.csv")

    action = "Writing"
    if os.path.exists(name):
        action = "Appending"
        results_out = open(name, "a+", newline='\n', encoding='utf-8')
    else:
        results_out = open(name, "w+", newline='\n', encoding='utf-8')
        results_out.write(f"File,AUC,ACC,NumPatterns,Filtering\n")


    results_out.write(f"{datasetName},{str(auc)}, {str(acc)}, {str(numPatterns)}, {str(filtering)}\n")

    results_out.close()

    return name

def GetFromFile(file):
    if os.path.isfile(file):
        filename, file_extension = os.path.splitext(file)
        if file_extension == ".arff":
            return ReadARFF(file)
        elif file_extension == ".dat":
            return ReadDAT(file)
        else:
            raise Exception(
                f"Extension '{file_extension}' of file '{filename}' is not supported ")
    else:
        raise Exception(f"File: {file} is not valid")

def returnX_y(file):
        arff_file = GetFromFile(file)
        instancesDf = get_dataframe_from_arff(arff_file)
        instances = instancesDf.to_numpy()
        X = instances[:, 0:len(instances[0])-1]
        y = instances[:, len(instances[0])-1 : len(instances[0])]
        return X,y

def get_dataframe_from_arff(arff_file):
    instancesDf = pd.DataFrame.from_records(
            arff_file['data'], columns=list(
                map(lambda attribute: attribute[0], arff_file['attributes'])
            )
        )
    return instancesDf



import math
import copy
from .Helpers import ArgMax


class BasicEvaluation(object):
    def __init__(self):
        self.TP = 0
        self.TN = 0
        self.FP = 0
        self.FN = 0

        self.TPrate = 0
        self.TNrate = 0
        self.FPrate = 0
        self.FNrate = 0

        self.specificity = 0
        self.sensitivity = 0
        self.precision = 0
        self.recall = 0
        self.Yrate = 0


class ConfusionMatrix(object):
    def __init__(self, classes):
        numclass = len(classes)
        self.Classes = classes
        self.__lettersposition = list(
            map(lambda x: str(chr(x+ord('a'))), list(range(26))))
        self.Matrix = [[0 for i in range(numclass)] for j in range(numclass+1)]

    def ComputeBasicEvaluation(self, positiveClass):
        basicEvaluation = BasicEvaluation()
        N = sum(map(sum, self.Matrix))
        basicEvaluation.TP = self.Matrix[positiveClass][positiveClass]

        for i in range(len(self.Matrix[0])):
            basicEvaluation.TN += self.Matrix[i][i]
            basicEvaluation.FP += self.Matrix[positiveClass][i]

        basicEvaluation.TN -= basicEvaluation.TP
        basicEvaluation.FP -= basicEvaluation.TP

        # to add abstentions
        basicEvaluation.FP += self.Matrix[len(self.Matrix) - 1][positiveClass]

        basicEvaluation.FN = N - \
            (basicEvaluation.TP + basicEvaluation.FP + basicEvaluation.TN)

        try:
            basicEvaluation.TPrate = basicEvaluation.TP * \
                1.0 / (basicEvaluation.TP + basicEvaluation.FN)
        except ZeroDivisionError:
            basicEvaluation.TPrate = 0.0
        try:
            basicEvaluation.TNrate = basicEvaluation.TN * \
                1.0 / (basicEvaluation.TN + basicEvaluation.FP)
        except ZeroDivisionError:
            basicEvaluation.TNrate = 0.0
        try:
            basicEvaluation.FPrate = basicEvaluation.FP * \
                1.0 / (basicEvaluation.TN + basicEvaluation.FP)
        except ZeroDivisionError:
            basicEvaluation.FPrate = 0.0
        try:
            basicEvaluation.FNrate = basicEvaluation.FN * \
                1.0 / (basicEvaluation.TP + basicEvaluation.FN)
        except ZeroDivisionError:
            basicEvaluation.FNrate = 0.0

        basicEvaluation.sensitivity = basicEvaluation.TPrate
        basicEvaluation.specificity = basicEvaluation.TNrate

        try:
            basicEvaluation.precision = basicEvaluation.TP * \
                1.0 / (basicEvaluation.TP + basicEvaluation.FP)
        except ZeroDivisionError:
            basicEvaluation.precision = 0

        basicEvaluation.recall = basicEvaluation.TPrate

        try:
            basicEvaluation.Yrate = (
                basicEvaluation.TP + basicEvaluation.FP) * 1.0 / N
        except ZeroDivisionError:
            basicEvaluation.Yrate = 0

        return basicEvaluation

    def AUCMeasure(self, positiveClass):
        basiceval = self.ComputeBasicEvaluation(positiveClass)
        return (1 + (basiceval.TPrate - basiceval.FPrate))/2

    def __repr__(self):
        lettersClasses = '|\t'.join(
            [self.__lettersposition[i] for i in range(len(self.Classes))])
        headers_row = '|\t'.join(
            [":--:" for i in range(len(self.Classes))])
        result = f"|\t{lettersClasses}|\t<-- classified as |\r\n|\t{headers_row}|\t---|\r\n"
        for row in range(len(self.Matrix)):
            result += "|\t"+'|\t'.join([str(self.Matrix[row][col])
                                        for col in range(len(self.Matrix[0]))])
            if row < len(self.Matrix)-1:
                result += f"\t| {self.__lettersposition[row]} = {self.Classes[row]}|\r\n"
            if row == len(self.Matrix)-1:
                result += f"\t| The last row correspond to the abstentions|\r\n"

        return result


class CrispAndPartitionEvaluation(object):
    def __init__(self):
        self.ConfusionMatrix = None


def Evaluate(classes, real, predicted):
    if len(real) != len(predicted):
        raise Exception(
            "Cannot evaluate classification. Real and Predicted counts are different.")
    numClasses = len(classes)
    evaluation = CrispAndPartitionEvaluation()
    evaluation.ConfusionMatrix = ConfusionMatrix(classes)
    confusion = [[0]*2 for i in enumerate(classes)]
    classified_as = 0
    error_count = 0

    for i in range(len(real)):
        if real[i] != predicted[i]:
            error_count = error_count + 1
        confusion[real[i]][predicted[i]] = confusion[real[i]][predicted[i]] + 1

    acc = 100.0 * (len(real) - error_count) / len(real)
    auc = obtainAUCMulticlass(confusion, len(classes))
        
    
    return confusion, acc, auc

def AddMatrices(cmA, cmB):
    if not cmA:
        return cmB
    if not cmB:
        return cmA

    if len(cmA) != len(cmB):
        raise Exception("Matrix missmatch")

    newMatrix = ConfusionMatrix(cmA.Classes)

    newMatrix.Matrix = [[cmA[i][j] + cmB[i][j]
                         for j in range(len(cmA[0]))] for i in range(len(cmA))]
    return newMatrix


def NormalizeVotes(values):
    result = [0]*len(values)
    argMax = ArgMax(values)

    if values[argMax] == 0:
        return result

    result[argMax] = 1
    return result

def __obtainAUCBinary(tp, tn, fp, fn):
        nPos = tp +fn
        nNeg = tn + fp
        recall = tp / nPos
        if tp == 0:
            recall = 0
        
        sensibility = tn/ nNeg
        if tn == 0:
            sensibility = 0
        
        return (recall + sensibility) / 2

def obtainAUCMulticlass(confusion, num_classes):
    sumVal = 0
    for i in range(num_classes):
        tp = confusion[i][i]

        for j in range(i+1, num_classes):
            fp = confusion[j][i]
            fn = confusion[i][j]
            tn = confusion[j][j]
            sumVal = sumVal + __obtainAUCBinary(tp, tn, fp, fn)
    
    avg = (sumVal * 2) / (num_classes * (num_classes-1))
    return avg

    




from copy import copy, deepcopy


class ForwardFeatureIterator(object):
    def __init__(self, dataset, features):
        self.__Dataset = dataset
        self.__CandidateFeatures = features
        self.__SelectedFeatures = list()
    
    @property
    def Dataset(self):
        return self.__Dataset
    @Dataset.setter
    def Dataset(self, new_dataset):
        self.__Dataset = new_dataset
    
    @property
    def CandidateFeatures(self):
        return self.__CandidateFeatures
    @CandidateFeatures.setter
    def CandidateFeatures(self, new_candidate_features):
        self.__CandidateFeatures = new_candidate_features

    @property
    def SelectedFeatures(self):
        return self.__Dataset
    @SelectedFeatures.setter
    def SelectedFeatures(self, new_selected_features):
        self.__SelectedFeatures = new_selected_features

    def Add(self, feature):
        if feature not in self.CandidateFeatures:
            return False
        self.SelectedFeatures.append(feature)
        self.CandidateFeatures.remove(feature)
        return True

    @property
    def FeaturesRemain(self):
        return len(self.CandidateFeatures) > 0

    def GetFeatures(self):
        result = list()
        for feature in self.CandidateFeatures:
            features = copy(self.SelectedFeatures)
            features.insert(0,feature)
            result.append(features)

        return result



import math


class DecisionTree():

    def __init__(self, dataset):
        self.__Model = dataset.Model
        self.__TreeRootNode = None
    
    @property
    def Model(self):
        return self.__Model

    @property
    def Size(self):
        if self.TreeRootNode is None:
            return 0
        else:
            return self.ComputeSizeTree(self.TreeRootNode)

    @property
    def Leaves(self):
        if self.TreeRootNode is None:
            return 0
        else:
            return self.ComputeLeaves(self.TreeRootNode)

    def ComputeSizeTree(self, decisionTree):
        if decisionTree.Children is None:
            return 1
        else:
            childrenSize = list(map(lambda child: self.ComputeSizeTree(child), decisionTree.Children))
            childrenSize.append(0)
            return max(childrenSize)+1

    def ComputeLeaves(self, decisionTree):
        if decisionTree.IsLeaf:
            return 1
        else:
            numLeaves = sum(map(lambda child: self.ComputeLeaves(child), decisionTree.Children))
            print(f"computeLeaves: {numLeaves} type {type(numLeaves)}")
            return numLeaves


class DecisionTreeNode():

    def __init__(self, data):
        self.__Data = data
        self.__Parent = None
        self.__ChildSelector = None
        self.__Children = []
    
    @property
    def Data(self):
        return self.__Data
    
    @property
    def Children(self):
        return self.__Children
    @Children.setter
    def Children(self, new_children):
        self.__Children = new_children

    @property
    def IsLeaf(self):
        return (not self.Children or len(self.Children) == 0)

    def __format__(self, ident):
        result = self.__repr__()

        if not self.IsLeaf:
            for child in range(len(self.Children)):
                if self.Children[child].Data:
                    childSelector = self.ChildSelector
                    curChild = self.Children[child]
                    result = f"{result}\n{' '*((ident+1)*3)}- {childSelector.__format__(child)} {curChild.__format__(ident+1)}"

        return result

    def __repr__(self):
        if not self.ChildSelector:
            return f"[{', '.join(map(str, self.Data))}]"
        else:
            return f"[{', '.join(map(str, self.Data))}] - {self.ChildSelector}"



import math
import numpy as np
from io import StringIO, BytesIO
from .Helpers import ArgMax, convert_to_ndarray
from .DecisionTreeBuilder import DecisionTreeBuilder, MultivariateDecisionTreeBuilder
from .PatternMiner import PatternMinerWithoutFiltering
from .DistributionEvaluator import Hellinger, MultiClassHellinger, QuinlanGain
from .DistributionTester import PureNodeStopCondition, AlwaysTrue
from .Item import SubsetRelation
from .Dataset import Dataset, FileDataset, PandasDataset
from .Evaluation import CrispAndPartitionEvaluation, Evaluate, obtainAUCMulticlass
from .PatternFilter import MaximalPatternsGlobalFilter
from tqdm import tqdm


class PBC4cip:

    def __init__(self, tree_count=100, filtering=False, multivariate = False, file_dataset = None):
        self.File = None
        self.__miner = None
        if filtering:
            filterer = MaximalPatternsGlobalFilter()
            self.__filterer = filterer
        else:
            self.__filterer = None
        self.__multivariate = multivariate
        self.__treeCount = tree_count
        if file_dataset is not None:
            self.__dataset = FileDataset(file_dataset)
        else:
            self.__dataset = None            
        self.__EmergingPatterns = list()
        self.__class_nominal_feature = None
        self.__normalizing_vector = None
        self.__votesSum = None
        self.__classDistribution = None

    @property
    def dataset(self):
        return self.__dataset

    @dataset.setter
    def dataset(self, new_dataset):
        self.__dataset = new_dataset

    @property
    def multivariate(self):
        return self.__multivariate

    @multivariate.setter
    def multivariate(self, new_multivariate):
        self.__multivariate = new_multivariate
    
    @property
    def miner(self):
        return self.__miner

    @miner.setter
    def miner(self, new_miner):
        self.__miner = new_miner

    @property
    def filterer(self):
        return self.__filterer

    @filterer.setter
    def filterer(self, new_filterer):
        self.__filterer = new_filterer
        
    @property
    def treeCount(self):
        return self.__treeCount

    @treeCount.setter
    def treeCount(self, new_treeCount):
        self.__treeCount = new_treeCount


    def fit(self, X, y):
        if self.dataset is None:
            self.dataset = PandasDataset(X,y)
            X = X.to_numpy()
            y = y.to_numpy()
            if  not isinstance(y[0], np.ndarray):
                y = convert_to_ndarray(y)    

        self.miner = PatternMinerWithoutFiltering()
        miner = self.miner
        miner.dataset = self.dataset
        miner.TreeCount = self.treeCount
        if self.multivariate:
            miner.decisionTreeBuilder = MultivariateDecisionTreeBuilder(self.dataset, X, y)
            miner.decisionTreeBuilder.distributionEvaluator = QuinlanGain
        else:
            miner.decisionTreeBuilder = DecisionTreeBuilder(self.dataset, X, y)
            miner.decisionTreeBuilder.distributionEvaluator = QuinlanGain
        self.EmergingPatterns = miner.Mine()
        if self.filterer is not None:
            self.EmergingPatterns = self.filterer.Filter(self.EmergingPatterns)
        self.__ComputeVotes(X, y, self.dataset.Class[1])
        return self.EmergingPatterns

    def __predict_inst(self, instance):
        votes = [0]*len(self._class_nominal_feature)

        for pattern in self.EmergingPatterns:
            if pattern.IsMatch(instance):
                for i in range(len(votes)):
                    votes[i] += pattern.Supports[i]

        result = [0]*(len(votes)) 
        for i,_ in enumerate(votes):
            try:
                result[i] = votes[i] * \
                    self._normalizing_vector[i] / self.__votesSum[i]
            except ZeroDivisionError:
                result[i] = 0

        if sum(result) > 0:
            return result
        else:
            return self.__classDistribution
    
    def predict(self, scored_samples):
        predicted = [ArgMax(instance) for instance in scored_samples]
        return predicted
    
    def score_samples(self, X):
        if isinstance(self.dataset, PandasDataset):
            X = X.to_numpy()

        classification_results = list()
        for instance in tqdm(X, desc=f"Classifying instances", unit="instance", leave=False):
            result = self.__predict_inst(instance)
            classification_results.append(result)

        return classification_results

    
    def __ComputeVotes(self, X, y, classes):
        self._class_nominal_feature = classes
        instancesByClass = self.__GroupInstancesByClass(
            X, y, classes)
        self._normalizing_vector = self.__ComputeNormalizingVector(
            instancesByClass, len(y))
        self.__classDistribution = self.__ComputeClassDistribution(
            instancesByClass,  len(y))

        self.__votesSum = [0]*len(classes)
        for pattern in self.EmergingPatterns:
            for i,_ in enumerate(classes):
                self.__votesSum[i] += pattern.Supports[i]

    def __GroupInstancesByClass(self, X, y, classes):
        instancesByClass = list()

        for _ in enumerate(classes):
            instancesByClass.append(list())

        for i,instance in enumerate(y):
            instancesByClass[self.dataset.GetClassValue(instance)].append(X[i])

        return instancesByClass

    def __ComputeNormalizingVector(self, instancesByClass, instanceCount):
        vectorSum = 0
        normalizingVector = [0]*len(instancesByClass)

        for i, _ in enumerate(instancesByClass):
            try:
                normalizingVector[i] = 1.0 - 1.0 * \
                    len(instancesByClass[i]) / \
                    instanceCount
            except ZeroDivisionError:
                normalizingVector[i] = 0
            vectorSum += normalizingVector[i]

        for i, _ in enumerate(normalizingVector):
            try:
                normalizingVector[i] /= vectorSum
            except ZeroDivisionError:
                normalizingVector[i] = 0

        return normalizingVector

    def __ComputeClassDistribution(self, instancesByClass, instanceCount):
        classDistribution = [0]*len(instancesByClass)
        for i, _ in enumerate(instancesByClass):
            try:
                classDistribution[i] = 1.0 * \
                    len(instancesByClass[i]) / \
                    instanceCount
            except ZeroDivisionError:
                classDistribution[i] = 0

        return classDistribution




from .Item import SubsetRelation
from .EmergingPatterns import EmergingPatternComparer
from .Item import ItemComparer
from tqdm import tqdm

class MaximalPatternsGlobalFilter(object):
    def __init__(self):
       self.__comparer = EmergingPatternComparer(ItemComparer)

    @property
    def comparer(self):
        return self.__comparer
    @comparer.setter
    def comparer(self, new_comparer):
        self.__comparer = new_comparer

    def Filter(self, patterns):
        selected_patterns = set()

        for candidate_pattern in tqdm(patterns, desc=f"Filtering the found patterns", unit="candidate_pattern", leave=False ):
            minimal_patterns = set()
            general_pattern_found = False

            for selected_pattern in selected_patterns:
                pattern_relation = self.comparer.Compare(candidate_pattern, selected_pattern)
                if pattern_relation == SubsetRelation.Subset or pattern_relation == SubsetRelation.Equal:
                    general_pattern_found = True
                    break
                elif pattern_relation == SubsetRelation.Superset:
                    minimal_patterns.add(selected_pattern)
            
            if not general_pattern_found:
                for minimal_pattern in minimal_patterns:
                    selected_patterns.remove(minimal_pattern)
                selected_patterns.add(candidate_pattern)
        
        return list(selected_patterns)


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PBC4cip",
    version="0.0.0.4",
    author="Ernesto Ramírez-Sáyago, Miguel Angel Medina-Pérez, Octavio Loyola-González",
    author_email="A00513925@itesm.mx, migue@tec.mx, octavioloyola@tec.mx",
    description="PBC4cip classifier for class imbalance problems",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/octavioloyola/PBC4cip",
    packages=setuptools.find_packages(),
    include_package_data = True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    install_requires=['numpy', 'pandas', 'scikit-learn', 'tqdm'],
    python_requires='>=3.6',
)



def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

import math
from .Helpers import FindDistribution, Substract
import operator
from .FeatureSelectors import CutPointSelector, MultipleValuesSelector, ValueAndComplementSelector, MultivariateCutPointSelector
from copy import copy, deepcopy
import numpy as np
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA

class SplitIteratorProvider(object):
    def __init__(self, dataset):
        self.Dataset = dataset

    def GetSplitIterator(self, feature):
        currentFeature = self.Dataset.GetAttribute(feature)
        if self.Dataset.IsNominalFeature(currentFeature):
            return NominalSplitIterator(self.Dataset, currentFeature)
        else:
            return NumericSplitIterator(self.Dataset, currentFeature)


class SplitIterator(object):
    def __init__(self, dataset, feature):
        self.__Dataset = dataset
        self.__Model = self.Dataset.Model
        self.__Class = self.Dataset.Class
        self.__Feature = feature
        self.__CurrentDistribution = None
        self.__initialized = False
        self.__numClasses = 0
        self.__instances = 0

    @property
    def Dataset(self):
        return self.__Dataset
    @Dataset.setter
    def Dataset(self, new_dataset):
        self.__Dataset = new_dataset
    
    @property
    def Model(self):
        return self.__Model
    @Model.setter
    def Model(self, new_model):
        self.__Model = new_model
    
    @property
    def Class(self):
        return self.__Class
    @Class.setter
    def Class(self, new_class):
        self.__Class = new_class
    
    @property
    def Feature(self):
        return self.__Feature
    @Feature.setter
    def Feature(self, new_feature):
        self.__Feature = new_feature

    def Initialize(self, instances):
        if not self.Model:
            raise Exception("Model is null")
        if self.Class[1] in ['numeric', 'real', 'integer', 'string']:
            raise Exception("Cannot use this iterator on non-nominal class")
        self._numClasses = len(self.Dataset.GetClasses())
        self._instances = len(instances)
        self.CurrentDistribution = [[] for i in range(2)]
        self._initialized = True

    def FindNext(self):
        if not self._initialized:
            raise Exception("Iterator not initialized")
        return False

    def CreateCurrentChildSelector(self):
        return None
    def GetFeatureIdx(self):
        return self.Dataset.GetFeatureIdx(self.Feature)

    def IsMissing(self, instance):
        return self.Dataset.IsMissing(self.Feature, instance)

    def GetFeatureValue(self, instance):
        return self.Dataset.GetFeatureValue(self.Feature, instance)

    def GetClassValue(self, instance):
        return self.Dataset.GetClasses().index(instance[self.Dataset.GetClassIdx()])


class NumericSplitIterator(SplitIterator):

    def __init__(self, dataset, feature):
        super().__init__(dataset, feature)
        self.__cuttingStrategy = None
        self.__currentIndex = 0
        self.__lastClassValue = 0
        self.__sortedInstances = None
        self.__selectorFeatureValue = 0

    def Initialize(self, instances):
        super().Initialize(instances)

        self._initialized = True

        if self.Dataset.IsNominalFeature(self.Feature):
            raise Exception("Cannot use this iterator on non-numeric feature")

        if self.Feature[1].lower() in ["integer"]:
            self.__cuttingStrategy = self.NumericOnPoint
        elif self.Feature[1].lower() in ["real", "numeric"]:
            self.__cuttingStrategy = self.NumericCenterBetweenPoints
        else:
            raise Exception(
                f"Feature type {self.Feature[1]} is not considered")

        instList = list(instances)
        self.__sortedInstances = list(
            filter(lambda element: not self.IsMissing(element[0]), instList))

        sortedInsts = sorted(instList, key=lambda element: element[0][self.GetFeatureIdx()])
        self.__sortedInstances = sortedInsts


        self.CurrentDistribution[0] = [0]*self._numClasses
        self.CurrentDistribution[1] = FindDistribution(
            self.__sortedInstances, self.Model, self.Dataset.Class)


        if (len(self.__sortedInstances) == 0):
            return

        self.__currentIndex = -1
        self.__lastClassValue = self.FindNextClass(0)

    def FindNext(self):
        super().FindNext()
        if (self.__currentIndex >= len(self.__sortedInstances) - 1):
            return False

        self.__currentIndex += 1
        while self.__currentIndex < len(self.__sortedInstances) - 1:
            instance = self.__sortedInstances[self.__currentIndex][0]
            objClass = self.GetClassValue(instance)
            self.CurrentDistribution[0][objClass] += self.__sortedInstances[self.__currentIndex][1]
            self.CurrentDistribution[1][objClass] -= self.__sortedInstances[self.__currentIndex][1]

            if self.GetFeatureValue(instance) != self.GetFeatureValue(self.__sortedInstances[self.__currentIndex+1][0]):
                nextClassValue = self.FindNextClass(self.__currentIndex + 1)
                if (self.__lastClassValue != nextClassValue) or (self.__lastClassValue == -1 and nextClassValue == -1):
                    self.__selectorFeatureValue = self.__cuttingStrategy(
                        instance)
                    self.__lastClassValue = nextClassValue
                    return True
            self.__currentIndex += 1
        return False

    def CreateCurrentChildSelector(self):
        selector = CutPointSelector(self.Dataset, self.Feature)
        selector.CutPoint = self.__selectorFeatureValue
        return selector

    def FindNextClass(self, index):
        currentClass = self.GetClassValue(self.__sortedInstances[index][0])
        currentValue = self.GetFeatureValue(self.__sortedInstances[index][0])
        index += 1
        while index < len(self.__sortedInstances) and currentValue == self.GetFeatureValue(self.__sortedInstances[index][0]):
            if currentClass != self.GetClassValue(self.__sortedInstances[index][0]):
                return -1
            index += 1
        return currentClass

    def NumericOnPoint(self, instance):
        return instance[self.GetFeatureIdx()]

    def NumericCenterBetweenPoints(self, instance):
        return (instance[self.GetFeatureIdx()] + self.__sortedInstances[self.__currentIndex+1][0][self.GetFeatureIdx()]) / 2


class NominalSplitIterator(SplitIterator):

    def __init__(self, dataset, feature):
        super().__init__(dataset, feature)
        self.__perValueDistribution = None
        self.__totalDistribution = None
        self.__valuesCount = None
        self.__existingValues = None
        self.__iteratingTwoValues = None
        self.__valueIndex = None
        self.__twoValuesIterated = None

    def Initialize(self, instances):
        super().Initialize(instances)
        self.__perValueDistribution = {}
        self.__totalDistribution = [0]*self._numClasses

        for instance in instances:
            if self.IsMissing(instance[0]):
                continue
            value = self.GetFeatureValue(instance[0])
            current = [0]*self._numClasses
            if not value in self.__perValueDistribution:
                self.__perValueDistribution.update({value: current})

            classIdx = self.GetClassValue(instance[0])
            self.__perValueDistribution[value][classIdx] += instance[1]
            self.__totalDistribution[classIdx] += instance[1]

        self.__valuesCount = len(self.__perValueDistribution)
        self.__existingValues = list(self.__perValueDistribution.keys())
        self.__iteratingTwoValues = (self.__valuesCount == 2)
        self.__valueIndex = -1
        self.__twoValuesIterated = False

    def FindNext(self):
        super().FindNext()
        if self.__valuesCount == self._instances:
            return False
        if self.__iteratingTwoValues:
            if self.__twoValuesIterated:
                return False
            self.__twoValuesIterated = True
            self.CalculateCurrent(
                self.__perValueDistribution[self.__existingValues[0]])
            return True
        else:
            if(self.__valuesCount < 2 or self.__valueIndex >= self.__valuesCount - 1):
                return False
            self.__valueIndex += 1
            self.CalculateCurrent(
                self.__perValueDistribution[self.__existingValues[self.__valueIndex]])
            return True

    def CreateCurrentChildSelector(self):
        if self.__iteratingTwoValues:
            selector = MultipleValuesSelector(self.Dataset, self.Feature)
            selector.Values = list(self.__perValueDistribution.keys())
        else:
            selector = ValueAndComplementSelector(self.Dataset, self.Feature)
            selector.Value = self.__existingValues[self.__valueIndex]
        return selector


    def CalculateCurrent(self, current):
        self.CurrentDistribution[0] = current
        self.CurrentDistribution[1] = Substract(
            self.__totalDistribution, current)

class MultivariateSplitIteratorProvider(SplitIteratorProvider):
    def __init__(self, dataset):
        self.Dataset = dataset

    def GetMultivariateSplitIterator(self, features, wMin):
        result = MultivariateOrderedFeatureSplitIterator(
            self.Dataset, features)
        result.WMin = wMin
        return result


class MultivariateSplitIterator(SplitIterator):
    def __init__(self, dataset, features):
        super().__init__(dataset, None)
        self.Features = features

    def Initialize(self, instances):
        raise Exception("Must initialize as multivariate")

    def InitializeMultivariate(self, instances, node):
        if not self.Model:
            raise Exception("Model is null")
        if self.Class[1] in ['numeric', 'real', 'integer', 'string']:
            raise Exception("Cannot use this iterator on non-nominal class")
        if any(self.Dataset.IsNominalFeature(feature) for feature in self.Features):
            raise Exception("Cannot use this iterator on numeric features")

        self._numClasses = len(self.Dataset.GetClasses())
        self._instances = len(instances)
        self.CurrentDistribution = [[] for i in range(2)]
        self._initialized = True

    def FindNext(self):
        super().FindNext()


class MultivariateOrderedFeatureSplitIterator(MultivariateSplitIterator):
    def __init__(self, dataset, features):
        super().__init__(dataset, features)
        self.__filteredInstances = None
        self.__projections = None
        self.__currentIndex = 0
        self.__lastClassValue = None
        self.__sortedInstances = None
        self.__cuttingStrategy = None
        self.__selectorFeatureValue = None
        self.__weights = None
        self.WMin = None

    def InitializeMultivariate(self, instances, node):
        super().InitializeMultivariate(instances, node)

        if not self.__cuttingStrategy:
            self.__cuttingStrategy = self.NumericOnPoint
        else:
            self.__cuttingStrategy = self.NumericCenterBetweenPoints

        self.__filteredInstances = list(
            filter(lambda instance: not any(self.Dataset.IsMissing(feature, instance[0]) for feature in self.Features), instances))

        self.__projections = self.GetProjections(self.__filteredInstances)

        if not self.__projections or len(self.__projections) == 0:
            return False

        self.__sortedInstances = list()

        self.__sortedInstances = [(self.__filteredInstances[i][0], self.__filteredInstances[i][1],
                                   self.__projections[i]) for i in range(len(self.__filteredInstances))]
        self.__sortedInstances.sort(key=lambda instance: instance[2])

        self.CurrentDistribution[0] = [0]*self._numClasses
        self.CurrentDistribution[1] = FindDistribution(
            self.__sortedInstances, self.Model, self.Dataset.Class)

        self.__currentIndex = -1
        self.__lastClassValue = self.FindNextClass(0)
        return True

    def FindNext(self):
        super().FindNext()
        if (self.__currentIndex >= len(self.__sortedInstances) - 1):
            return False

        self.__currentIndex += 1
        while self.__currentIndex < len(self.__sortedInstances) - 1:
            instance = self.__sortedInstances[self.__currentIndex][0]
            value = self.__sortedInstances[self.__currentIndex][2]
            objClass = self.GetClassValue(instance)

            self.CurrentDistribution[0][objClass] += self.__sortedInstances[self.__currentIndex][1]
            self.CurrentDistribution[1][objClass] -= self.__sortedInstances[self.__currentIndex][1]

            if value != self.__sortedInstances[self.__currentIndex+1][2]:
                nextClassValue = self.FindNextClass(self.__currentIndex + 1)
                if (self.__lastClassValue != nextClassValue) or (self.__lastClassValue == -1 and nextClassValue == -1):
                    self.__selectorFeatureValue = self.__cuttingStrategy(value)
                    self.__lastClassValue = nextClassValue
                    return True
            self.__currentIndex += 1
        return False

    def CreateCurrentChildSelector(self):
        selector = MultivariateCutPointSelector(self.Dataset, self.Features)
        selector.CutPoint = self.__selectorFeatureValue
        selector.Weights = self.__weights
        return selector

    def FindNextClass(self, index):
        currentClass = self.GetClassValue(self.__sortedInstances[index][0])
        currentValue = self.__sortedInstances[index][2]
        index += 1
        while index < len(self.__sortedInstances) and currentValue == self.__sortedInstances[index][2]:
            if currentClass != self.GetClassValue(self.__sortedInstances[index][0]):
                return -1
            index += 1
        return currentClass

    def GetProjections(self, instances):
        classIdx = self.Dataset.GetClassIdx()
        featuresIdxs = [self.Dataset.GetFeatureIdx(
            feature) for feature in self.Features]
        ldaData = [[instance[0][featureIdx] for featureIdx in featuresIdxs]
                   for instance in instances]
        ldaTargets = [self.Dataset.GetIndexOfValue(
            self.Class[0], instance[0][classIdx]) for instance in instances]

        lda = LDA(n_components=1)
        try:
            ldaOutput = lda.fit(ldaData, ldaTargets).transform(ldaData)

            if len(ldaOutput) == 0:
                return list()

            w = lda.coef_[0]
            if len(w) == 0:
                return list()

            self.__weights = {self.Features[i]: w[i]
                              for i in range(0, len(self.Features))}

            w_norm = math.sqrt(sum(map(lambda x: math.pow(x, 2), w)))

            for x in w:
                if (w_norm == 0):
                    print(f"x: {x} w_norm {w_norm}")
                
                divNum = abs(x/w_norm)
                if ( (not math.isnan(divNum)) and divNum < self.WMin):
                    print("x/w_norm is smaller than wMin")
                    return list()

            return list(map(lambda r: r[0], ldaOutput))

        except Exception as e:
            return list()

    def NumericOnPoint(self, value):
        return value

    def NumericCenterBetweenPoints(self, value):
        return (value + self.__sortedInstances[self.__currentIndex][2]) / 2



import sys
import math
import copy
from _functools import cmp_to_key


class WinningSplitSelector:

    def __init__(self, whichBetterToFind=None):
        if not whichBetterToFind or whichBetterToFind <= 0:
            raise Exception("WhichBetterToFind must be positive")
        else:
            self.__whichBetterToFind = whichBetterToFind
        self.MinStoredValue = sys.float_info.min
        self.List = list(tuple())

    @property
    def WinningSelector(self):
        index = min(self.__whichBetterToFind - 1, len(self.List) - 1)
        return self.List[index][1]

    @property
    def WinningDistribution(self):
        index = min(self.__whichBetterToFind - 1, len(self.List) - 1)
        return self.List[index][2]

    def EvaluateThis(self, currentGain, splitIterator, level):

        if (len(self.List) < self.__whichBetterToFind or currentGain > self.MinStoredValue):
            currentChildSelector = splitIterator.CreateCurrentChildSelector()
            copyOfCurrentDistribution = copy.deepcopy(
                splitIterator.CurrentDistribution)
            self.List.append(
                tuple((currentGain, currentChildSelector, copyOfCurrentDistribution)))
            self.List.sort(key=cmp_to_key(Compare))

            if len(self.List) > self.__whichBetterToFind:
                self.List.remove(self.List[self.__whichBetterToFind])
            index = min(self.__whichBetterToFind-1, len(self.List)-1)

            self.MinStoredValue = self.List[index][0]
            return True
        return False

    def IsWinner(self):
        return len(self.List) > 0


def Compare(x, y):
    val = x[0]-y[0]
    if val > 0:
        return -1
    elif val < 0:
        return 1
    else:
        return 0



import random
import math
import pandas as pd
from tqdm import tqdm
from multiprocessing import Pool, freeze_support, cpu_count
from functools import partial
from .SupervisedClassifier import DecisionTreeClassifier
from .RandomSampler import SampleWithoutRepetition
from .EmergingPatterns import EmergingPatternCreator, EmergingPatternComparer, EmergingPatternSimplifier
from .DistributionTester import AlwaysTrue
from .Item import ItemComparer, SubsetRelation
from .FilteredCollection import FilteredCollection


class PatternMinerWithoutFiltering:

    def __init__(self, treeCount=None, featureCount=None, minePatternsWhileBuildingTree=None):
        self.__dataset = None
        self.__Patterns = list()
        self.__PatternsList = list()
        self.__decisionTreeBuilder = None
        self.__EPTester = AlwaysTrue
        self.__FilterRelation = SubsetRelation.Superset

        self.__emergingPatternCreator = None
        self.__emergingPatternComparer = None
        self.__emergingPatternSimplifier = None
        self.__minimal = None

        if not featureCount:
            self.__FeatureCount = -1
        else:
            self.__FeatureCount = featureCount

        if not treeCount:
            self.__TreeCount = 100
        else:
            self.__TreeCount = treeCount

        if not minePatternsWhileBuildingTree:
            self.__MinePatternsWhileBuildingTree = False
        else:
            self.__MinePatternsWhileBuildingTree = minePatternsWhileBuildingTree

    @property
    def TreeCount(self):
        return self.__TreeCount
    @TreeCount.setter
    def TreeCount(self, new_tree_count):
        self.__TreeCount = new_tree_count

    @property
    def dataset(self):
        return self.__dataset
    @dataset.setter
    def dataset(self, new_dataset):
        self.__dataset = new_dataset
    
    @property
    def decisionTreeBuilder(self):
        return self.__decisionTreeBuilder
    @decisionTreeBuilder.setter
    def decisionTreeBuilder(self, new_dtb):
        self.__decisionTreeBuilder = new_dtb

    @property
    def FeatureCount(self):
        return self.__FeatureCount
    @FeatureCount.setter
    def FeatureCount(self, new_Feature_Count):
        self.__FeatureCount = new_Feature_Count
    
    @property
    def EPTester(self):
        return self.__EPTester

    @property
    def PatternsList(self):
        return self.__PatternsList

    @PatternsList.setter
    def PatternsList(self, new_patterns_list):
        self.__PatternsList = new_patterns_list

    def Mine(self):
        self.Patterns = list()
        self.__emergingPatternCreator = EmergingPatternCreator(self.dataset)
        self.__emergingPatternComparer = EmergingPatternComparer(
            ItemComparer().Compare)
        self.__emergingPatternSimplifier = EmergingPatternSimplifier(
            ItemComparer().Compare)
        self.Patterns = self.__DoMine(
            self.__emergingPatternCreator, self.PatternFound)
        self.PatternsList = []

        return self.Patterns

    def __DoMine(self, emergingPatternCreator, action):
        freeze_support()  # for Windows support
        featureCount = 0
        if self.FeatureCount != -1:
            featureCount = self.FeatureCount
        else:
            featureCount = int(math.log(len(self.dataset.Attributes), 2) + 1)

        decision_tree_builder = self.decisionTreeBuilder
        decision_tree_builder.FeatureCount = featureCount
        decision_tree_builder.OnSelectingFeaturesToConsider = SampleWithoutRepetition
        for i in tqdm(range(self.TreeCount), unit="tree", desc="Building trees and extracting patterns", leave=False):
            decision_tree_builder.OnSelectingFeaturesToConsider = SampleWithoutRepetition
            tree = decision_tree_builder.Build()
            treeClassifier = DecisionTreeClassifier(tree)
            emergingPatternCreator.ExtractPatterns(treeClassifier, action)

        return self.PatternsList

    def PatternFound(self, pattern):        
        if self.EPTester(pattern.Counts, self.dataset.Model, self.dataset.Class):
            simplifiedPattern = self.__emergingPatternSimplifier.Simplify(pattern)
            self.PatternsList.append(simplifiedPattern)
                